import os
import numpy as np
import re
from getdist.paramnames import ParamNames, ParamInfo, escapeLatex
from getdist.convolve import autoConvolve
from getdist import cobaya_interface
import pickle
import logging
from copy import deepcopy
from collections import namedtuple
from typing import Sequence, Any, Optional, Union, List

# whether to write to terminal chain names and burn in details when loaded from file
print_load_details = True

_int_types = (int, np.integer)
ParamConfidenceData = namedtuple("ParamConfidenceData", ("paramVec", "norm", "indexes", "cumsum"))

try:
    import pandas
except ImportError:
    pandas = None


class WeightedSampleError(Exception):
    """
    An exception that is raised when a WeightedSamples error occurs
    """


class ParamError(WeightedSampleError):
    """
    An Exception that indicates a bad parameter.
    """


def print_load_line(message):
    if print_load_details:
        print(message)


def last_modified(files):
    """
    Returns the the latest "last modified" time for the given list of files. Ignores files that do not exist.

    :param files: An iterable of file names.
    :return: The latest "last modified" time
    """
    return max(os.path.getmtime(fname) for fname in files if os.path.exists(fname))


def slice_or_none(x, start=None, end=None):
    return getattr(x, "__getitem__", lambda _: None)(slice(start, end))


def findChainFileRoot(chain_dir, root, search_subdirectories=True):
    """
    Finds chain files with name root somewhere under chain_dir directory tree.
    root can also be a relative path relaqtive to chain_dir, or have leading directories as needed to make unique

    :param chain_dir: root directory of hierarchy of directories to look in
    :param root: root name for the chain
    :param search_subdirectories: recursively look in subdirectories under chain_dir
    :return: full path and root if found, otherwise None
    """
    root = root.replace('/', os.sep).replace('\\', os.sep)
    file_root = os.path.join(chain_dir, root)
    if hasChainFiles(file_root):
        return file_root
    if search_subdirectories:
        for base, dirs, files in os.walk(chain_dir):
            for _dir in dirs:
                file_root = os.path.join(base, _dir, root)
                if hasChainFiles(file_root):
                    return file_root
    return None


def chainFiles(root, chain_indices=None, ext='.txt', separator="_",
               first_chain=0, last_chain=-1, chain_exclude=None):
    """
    Creates a list of file names for samples given a root name and optional filters

    :param root: Root name for files (no extension)
    :param chain_indices: If True, only indexes inside the list included, If False, includes all indexes.
    :param ext: extension for files
    :param separator: separator character used to indicate chain number (usually _ or .)
    :param first_chain: The first index to include.
    :param last_chain: The last index to include.
    :param chain_exclude: A list of indexes to exclude, None to include all
    :return: The list of file names
    """

    folder = os.path.dirname(root)
    if root.endswith((os.sep, "/")):
        reg_exp = re.compile('(?P<num>[0-9]+)?' + re.escape(ext))
    else:
        basename = os.path.basename(root)
        reg_exp = re.compile(re.escape(basename) + '(' + re.escape(separator) + '(?P<num>[0-9]+))?' + re.escape(ext))
    files = []
    for f in os.listdir(folder):
        match = reg_exp.fullmatch(f)
        if match:
            index = int(match.group("num") or 0)
            if (chain_indices is None or index in chain_indices) \
                    and (chain_exclude is None or index not in chain_exclude) \
                    and index >= first_chain and (last_chain < 0 or index <= last_chain):
                files.append(os.path.join(folder, f))
    return files


def hasChainFiles(file_root, ext='.txt'):
    return any(chainFiles(file_root, ext=ext, separator=sep, last_chain=1) for sep in ['_', '.'])


_pandas_suggestion = True


def loadNumpyTxt(fname, skiprows=None):
    """
    Utility routine to loads numpy array from file.
    Uses faster pandas read routine if pandas is installed, or falls back to numpy's loadtxt otherwise

    :param fname: The file to load
    :param skiprows: The number of rows to skip at the begging of the file
    :return: numpy array of the data values
    """
    try:
        if pandas:
            return pandas.read_csv(fname, delim_whitespace=True, header=None, dtype=np.float64,
                                   skiprows=skiprows, comment='#').values
        else:
            global _pandas_suggestion
            if _pandas_suggestion:
                _pandas_suggestion = False
                logging.warning('Install pandas for faster reading from text files')
            return np.atleast_2d(np.loadtxt(fname, skiprows=skiprows or 0))
    except ValueError:
        print('Error reading %s' % fname)
        raise


def getSignalToNoise(C, noise=None, R=None, eigs_only=False):
    """
    Returns w, M, where w is the eigenvalues of the signal to noise (small y better constrained)

    :param C: covariance matrix
    :param noise: noise matrix
    :param R: rotation matrix, defaults to inverse of Cholesky root of the noise matrix
    :param eigs_only: only return eigenvalues
    :return: eigenvalues and matrix
    """
    if R is None:
        if noise is None:
            raise WeightedSampleError('Must give noise or rotation R')
        R = np.linalg.inv(np.linalg.cholesky(noise))

    M = np.dot(R, C).dot(R.T)
    if eigs_only:
        return np.linalg.eigvalsh(M)
    else:
        w, U = np.linalg.eigh(M)
        U = np.dot(U.T, R)
        return w, U


def covToCorr(cov, copy=True):
    """
    Convert covariance matrix to correlation matrix

    :param cov: The covariance matrix to work on
    :param copy: True if we shouldn't modify the input matrix, False otherwise.
    :return: correlation matrix
    """
    if copy:
        cov = cov.copy()
    for i, di in enumerate(np.sqrt(cov.diagonal())):
        if di:
            cov[i, :] /= di
            cov[:, i] /= di
    return cov


class ParSamples:
    """
    An object used as a container for named parameter sample arrays
    """
    pass


# noinspection PyAttributeOutsideInit
class WeightedSamples:
    """
    WeightedSamples is the base class for a set of weighted parameter samples

    :ivar weights:  array of weights for each sample (default: array of 1)
    :ivar loglikes: array of -log(Likelihoods) for each sample (default: array of 0)
    :ivar samples: n_samples x n_parameters numpy array of parameter values
    :ivar n: number of parameters
    :ivar numrows: number of samples positions (rows in the samples array)
    :ivar name_tag: name tag for the samples
    """

    weights: Optional[np.ndarray]
    loglikes: Optional[np.ndarray]
    samples: Optional[np.ndarray]
    norm: Any
    n: Any
    numrows: Any
    paramNames: Optional[ParamNames]

    def __init__(self, filename=None, ignore_rows=0, samples=None, weights=None, loglikes=None, name_tag=None,
                 label=None, files_are_chains=True, min_weight_ratio=1e-30):
        """
        :param filename: A filename of a plain text file to load from
        :param ignore_rows:
            - if int >=1: The number of rows to skip at the file in the beginning of the file
            - if float <1: The fraction of rows to skip at the beginning of the file
        :param samples: array of parameter values for each sample, passed to :func:`setSamples`
        :param weights: array of weights
        :param loglikes: array of -log(Likelihood)
        :param name_tag: The name of this instance.
        :param label: latex label for these samples
        :param files_are_chains: use False if the samples file (filename) does not start with two columns giving
                                 weights and -log(Likelihoods)
        :param min_weight_ratio: remove samples with weight less than min_weight_ratio times the maximum weight
        """

        self.precision = '%.8e'
        self.min_weight_ratio = min_weight_ratio
        if filename:
            cols = loadNumpyTxt(filename, skiprows=ignore_rows)
            if not len(cols):
                raise WeightedSampleError('Empty chain: %s' % filename)
            self.setColData(cols, are_chains=files_are_chains)
            self.name_tag = name_tag or os.path.basename(filename)
        else:
            self.setSamples(slice_or_none(samples, ignore_rows),
                            slice_or_none(weights, ignore_rows),
                            slice_or_none(loglikes, ignore_rows))
            self.name_tag = name_tag
            if samples is not None and int(ignore_rows):
                if print_load_details:
                    print('Removed %s lines as burn in' % ignore_rows)
        self.label = label
        self.needs_update = True

    def setColData(self, coldata, are_chains=True):
        """
        Set the samples given an array loaded from file

        :param coldata: The array with columns of [weights, -log(Likelihoods)] and sample parameter values
        :param are_chains: True if coldata starts with two columns giving weight and -log(Likelihood)
        """
        if are_chains:
            self.setSamples(coldata[:, 2:], coldata[:, 0], coldata[:, 1])
        else:
            self.setSamples(coldata)

    def getLabel(self):
        """
        Return the latex label for the samples

        :return: the label
        """
        return self.label or escapeLatex(self.getName())

    def getName(self):
        """
        Returns the name tag of these samples.

        :return: The name tag
        """
        return self.name_tag

    def setSamples(self, samples, weights=None, loglikes=None, min_weight_ratio=None):
        """
        Sets the samples from numpy arrays

        :param samples: The samples values, n_samples x n_parameters numpy array, or can be a list of parameter vectors
        :param weights: Array of weights for each sample. Defaults to 1 for all samples if unspecified.
        :param loglikes: Array of -log(Likelihood) values for each sample
        :param min_weight_ratio: remove samples with weight less than min_weight_ratio of the maximum
        """
        self.weights = weights
        self.loglikes = loglikes
        self.samples = samples
        if samples is not None:
            if isinstance(samples, (list, tuple)):
                samples = np.hstack([x.reshape(-1, 1) for x in samples])
            elif len(samples.shape) == 1:
                samples = np.atleast_2d(samples).transpose()
            self.samples = samples
            self.n = self.samples.shape[1]
            self.numrows = self.samples.shape[0]
            if min_weight_ratio is None:
                min_weight_ratio = self.min_weight_ratio
            if min_weight_ratio is not None and min_weight_ratio >= 0:
                self.setMinWeightRatio(min_weight_ratio)
        self._weightsChanged()

    def changeSamples(self, samples):
        """
        Sets the samples without changing weights and loglikes.

        :param samples: The samples to set
        """
        self.setSamples(samples, self.weights, self.loglikes)

    def _weightsChanged(self):
        if self.weights is not None:
            self.norm = np.sum(self.weights)
        elif self.samples is not None:
            self.weights = np.ones(self.numrows)
            self.norm = np.float64(self.numrows)
        self.means = None
        self.mean_loglike = None
        self.diffs = None
        self.fullcov = None
        self.correlationMatrix = None
        self.vars = None
        self.sddev = None
        self.needs_update = True

    def _makeParamvec(self, par):
        if isinstance(par, _int_types):
            if 0 <= par < self.n:
                return self.samples[:, par]
            elif par == -1:
                if self.loglikes is None:
                    raise WeightedSampleError('Samples do not have logLikes (par=-1)' % par)
                return self.loglikes
            elif par == -2:
                return self.weights
            else:
                raise WeightedSampleError('Parameter %i does not exist' % par)
        return par

    def getCov(self, nparam=None, pars=None):
        """
        Get covariance matrix of the parameters. By default uses all parameters, or can limit to max number or list.

        :param nparam: if specified, only use the first nparam parameters
        :param pars: if specified, a list of parameter indices (0,1,2..) to include
        :return: covariance matrix.
        """
        if self.fullcov is None:
            self._setCov()
        if pars is not None:
            return self.fullcov[np.ix_(pars, pars)]
        else:
            return self.fullcov[:nparam, :nparam]

    def _setCov(self):
        """
        Calculate and save the full covariance.

        :return: The full covariance matrix
        """
        self.fullcov = self.cov()
        return self.fullcov

    def getCorrelationMatrix(self):
        """
        Get the correlation matrix of all parameters

        :return: The correlation matrix
        """
        if self.correlationMatrix is None:
            self.correlationMatrix = covToCorr(self.getCov())
        return self.correlationMatrix

    def setMeans(self):
        """
        Calculates and saves the means of the samples

        :return: numpy array of parameter means
        """
        self.means = self.weights.dot(self.samples) / self.norm
        if self.loglikes is not None:
            self.mean_loglike = self.weights.dot(self.loglikes) / self.norm
        else:
            self.mean_loglike = None
        return self.means

    def getMeans(self, pars=None):
        """
        Gets the parameter means, from saved array if previously calculated.

        :param pars: optional list of parameter indices to return means for
        :return: numpy array of parameter means
        """
        if self.means is None:
            self.setMeans()
        if pars is None:
            return self.means
        else:
            return np.array([self.means[i] for i in pars])

    def getVars(self):
        """
        Get the parameter variances

        :return: A numpy array of variances.
        """
        if self.means is None:
            self.setMeans()
        self.vars = np.empty(self.n)
        for i in range(self.n):
            self.vars[i] = self.weights.dot((self.samples[:, i] - self.means[i]) ** 2) / self.norm
        self.sddev = np.sqrt(self.vars)
        return self.vars

    def setDiffs(self):
        """
        saves self.diffs array of parameter differences from the y, e.g. to later calculate variances etc.

        :return: array of differences
        """
        self.diffs = self.mean_diffs()
        return self.diffs

    def getAutocorrelation(self, paramVec, maxOff=None, weight_units=True, normalized=True):
        """
        Gets auto-correlation of an array of parameter values (e.g. for correlated samples from MCMC)

        By default uses weight units (i.e. standard units for separate samples from original chain).
        If samples are made from multiple chains, neglects edge effects.

        :param paramVec: an array of parameter values, or the int index of the parameter in stored samples to use
        :param maxOff: maximum autocorrelation distance to return
        :param weight_units: False to get result in sample point (row) units; weight_units=False gives standard
                             definition for raw chains
        :param normalized: Set to False to get covariance
                          (note even if normalized, corr[0]<>1 in general unless weights are unity).
        :return: zero-based array giving auto-correlations
        """
        if maxOff is None:
            maxOff = self.n - 1
        d = self.mean_diff(paramVec) * self.weights
        corr = autoConvolve(d, n=maxOff + 1, normalize=True)
        if normalized:
            corr /= self.var(paramVec)
        if weight_units:
            return corr * d.size / self.get_norm()
        else:
            return corr

    def getCorrelationLength(self, j, weight_units=True, min_corr=0.05, corr=None):
        """
        Gets the auto-correlation length for parameter j

        :param j: The index of the parameter to use
        :param weight_units: False to get result in sample point (row) units; weight_units=False gives standard
                             definition for raw chains
        :param min_corr: specifies a minimum value of the autocorrelation to use, e.g. where sampling noise is
                         typically as large as the calculation
        :param corr: The auto-correlation array to use, calculated internally by default
                     using :func:`getAutocorrelation`
        :return: the auto-correlation length
        """
        if corr is None:
            corr = self.getAutocorrelation(j, self.numrows // 10, weight_units=weight_units)
        ix = np.argmin(corr > min_corr * corr[0])
        N = corr[0] + 2 * np.sum(corr[1:ix])
        return N

    def getEffectiveSamples(self, j=0, min_corr=0.05):
        """
        Gets effective number of samples N_eff so that the error on mean of parameter j is sigma_j/N_eff

        :param j: The index of the param to use.
        :param min_corr: the minimum value of the auto-correlation to use when estimating the correlation length
        """
        return self.get_norm() / self.getCorrelationLength(j, min_corr=min_corr)

    def getEffectiveSamplesGaussianKDE(self, paramVec, h=0.2, scale=None, maxoff=None, min_corr=0.05):
        """
        Roughly estimate an effective sample number for use in the leading term for the MISE
        (mean integrated squared error) of a Gaussian-kernel KDE (Kernel Density Estimate). This is used for
        optimizing the kernel bandwidth, and though approximate should be better than entirely ignoring sample
        correlations, or only counting distinct samples.

        Uses fiducial assumed kernel scale h; result does depend on this (typically by factors O(2))

        For bias-corrected KDE only need very rough estimate to use in rule of thumb for bandwidth.

        In the limit h-> 0 (but still >0) answer should be correct (then just includes MCMC rejection duplicates).
        In reality correct result for practical h should depends on shape of the correlation function.

        If self.sampler is 'nested' or 'uncorrelated' return result for uncorrelated samples.

        :param paramVec: parameter array, or int index of parameter to use
        :param h: fiducial assumed kernel scale.
        :param scale: a scale parameter to determine fiducial kernel width, by default the parameter standard deviation
        :param maxoff: maximum value of auto-correlation length to use
        :param min_corr: ignore correlations smaller than this auto-correlation
        :return: A very rough effective sample number for leading term for the MISE of a Gaussian KDE.
        """
        if getattr(self, "sampler", "") in ["nested", "uncorrelated"]:
            return self.get_norm() ** 2 / np.dot(self.weights, self.weights)
        d = self._makeParamvec(paramVec)
        # Result does depend on kernel width, but hopefully not strongly around typical values ~ sigma/4
        kernel_std = (scale or self.std(d)) * h
        # Dependence is from very correlated points due to MCMC rejections;
        # Shouldn't need more than about correlation length
        if maxoff is None:
            maxoff = int(self.getCorrelationLength(d, weight_units=False) * 1.5) + 4
        maxoff = min(maxoff, self.numrows // 10)  # can get problems otherwise if weights are all very large
        uncorr_len = self.numrows // 2
        uncorr_term = 0
        nav = 0
        # first get expected value of each term for uncorrelated samples
        for k in range(uncorr_len, uncorr_len + 5):
            nav += self.numrows - k
            diff2 = (d[:-k] - d[k:]) ** 2 / kernel_std ** 2
            uncorr_term += np.dot(np.exp(-diff2 / 4) * self.weights[:-k], self.weights[k:])
        uncorr_term /= nav

        corr = np.zeros(maxoff + 1)
        corr[0] = np.dot(self.weights, self.weights)
        n = float(self.numrows)

        # Unoptimized calculation, slow for long correlation lengths
        # for k in range(1, maxoff + 1):
        #     diff2 = (d[:-k] - d[k:]) ** 2 / kernel_std ** 2
        #     corr[k] = np.dot(np.exp(-diff2 / 4) * self.weights[:-k], self.weights[k:]) - (n - k) * uncorr_term
        #     if corr[k] < min_corr * corr[0]:
        #         corr[k] = 0
        #         break
        # N1 = corr[0] + 2 * np.sum(corr[1:])

        def corr_k(_k):
            return np.dot(np.exp(-(d[:-_k] - d[_k:]) ** 2 / (4 * kernel_std ** 2)) * self.weights[:-_k],
                          self.weights[_k:]) - (n - _k) * uncorr_term

        threshold = min_corr * corr[0]
        corr[1] = corr_k(1)
        if corr[1] < threshold:
            N = corr[0]
        else:
            corr[2] = corr_k(2)
            if corr[2] > threshold:
                max_k = maxoff
                # for large correlation lengths, for speed need to sample rather than doing every k
                # find largest step for which correlation above threshold
                while max_k > 10:
                    test_val = corr_k(max_k // 3)
                    if test_val >= threshold:
                        break
                    max_k //= 3

                # does not need to be accurate
                if max_k < 20:
                    step_size = 1
                else:
                    step_size = max_k // 10

                cum_sum = corr[1] + corr[2]
                for k in range(3, maxoff + 1, step_size):
                    test_val = corr_k(k)
                    if test_val < threshold:
                        break
                    if k > 3:
                        cum_sum += test_val * step_size
                    else:
                        cum_sum += (test_val * step_size) / 2
                N = corr[0] + 2 * cum_sum
            else:
                N = corr[0] + 2 * corr[1]

        return self.get_norm() ** 2 / N

    def getEffectiveSamplesGaussianKDE_2d(self, i, j, h=0.3, maxoff=None, min_corr=0.05):
        """
        Roughly estimate an effective sample number for use in the leading term for the 2D MISE.
        If self.sampler is 'nested' or 'uncorrelated' return result for uncorrelated samples.

        :param i: parameter array, or int index of first parameter to use
        :param j: parameter array, or int index of second parameter to use
        :param h: fiducial assumed kernel scale.
        :param maxoff: maximum value of auto-correlation length to use
        :param min_corr: ignore correlations smaller than this auto-correlation
        :return: A very rough effective sample number for leading term for the MISE of a Gaussian KDE.
        """
        if getattr(self, "sampler", "") in ["nested", "uncorrelated"]:
            return self.get_norm() ** 2 / np.dot(self.weights, self.weights)
        d1 = self._makeParamvec(i)
        d2 = self._makeParamvec(j)
        cov = self.cov([d1, d2])
        if abs(cov[0, 1]) > np.sqrt(cov[0, 0] * cov[1, 1]) * 0.999:
            # totally correlated, fall back to 1D
            return self.getEffectiveSamplesGaussianKDE(i, h=h, min_corr=min_corr)
        # result does depend on kernel width, use fiducial h
        kernel_inv = np.linalg.inv(cov) / h ** 2

        # Dependence is from very correlated points due to MCMC rejections;
        # Shouldn't need more than about correlation length
        if maxoff is None:
            maxoff = int(max(self.getCorrelationLength(d1, weight_units=False),
                             self.getCorrelationLength(d2, weight_units=False)) * 1.5) + 4
        maxoff = min(maxoff, self.numrows // 10)  # can get problems otherwise if weights are all very large
        uncorr_len = self.numrows // 2
        uncorr_term = 0
        nav = 0
        # first get expected value of each term for uncorrelated samples
        for k in range(uncorr_len, uncorr_len + 5):
            nav += self.numrows - k
            delta = np.vstack((d1[:-k] - d1[k:], d2[:-k] - d2[k:]))
            diff2 = np.sum(delta * kernel_inv.dot(delta), 0)
            uncorr_term += np.dot(np.exp(-diff2 / 4) * self.weights[:-k], self.weights[k:])
        uncorr_term /= nav

        corr = np.zeros(maxoff + 1)
        corr[0] = np.dot(self.weights, self.weights)
        n = float(self.numrows)
        for k in range(1, maxoff + 1):
            delta = np.vstack((d1[:-k] - d1[k:], d2[:-k] - d2[k:]))
            diff2 = np.sum(delta * kernel_inv.dot(delta), 0)
            corr[k] = np.dot(np.exp(-diff2 / 4) * self.weights[:-k], self.weights[k:]) - (n - k) * uncorr_term
            if corr[k] < min_corr * corr[0]:
                corr[k] = 0
                break
        N = corr[0] + 2 * np.sum(corr[1:])
        return self.get_norm() ** 2 / N

    def weighted_sum(self, paramVec, where=None):
        """
        Calculates the weighted sum of a parameter vector, sum_i w_i p_i

        :param paramVec: array of parameter values or int index of parameter to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: weighted sum
        """
        paramVec = self._makeParamvec(paramVec)
        if where is None:
            return self.weights.dot(paramVec)
        return np.dot(paramVec[where], self.weights[where])

    def get_norm(self, where=None):
        """
        gets the normalization, the sum of the sample weights: sum_i w_i

        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: normalization
        """
        if where is None:
            if self.norm is None:
                self.norm = np.sum(self.weights)
            return self.norm
        else:
            return np.sum(self.weights[where])

    def mean(self, paramVec, where=None):
        """
        Get the mean of the given parameter vector.

        :param paramVec: array of parameter values or int index of parameter to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: parameter mean
        """
        if isinstance(paramVec, (list, tuple)):
            return np.array([self.weighted_sum(p, where) for p in paramVec]) / self.get_norm(where)
        else:
            return self.weighted_sum(paramVec, where) / self.get_norm(where)

    def var(self, paramVec, where=None):
        """
        Get the variance of the given parameter vector.

        :param paramVec: array of parameter values or int index of parameter to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: parameter variance
        """
        if isinstance(paramVec, (list, tuple)):
            return np.array([self.var(p) for p in paramVec])
        if where is not None:
            return np.dot(self.mean_diff(paramVec, where) ** 2, self.weights[where]) / self.get_norm(where)
        else:
            return np.dot(self.mean_diff(paramVec) ** 2, self.weights) / self.get_norm()

    def std(self, paramVec, where=None):
        """
        Get the standard deviation of the given parameter vector.

        :param paramVec: array of parameter values or int index of parameter to use
        :param where: if specified, a filter for the samples to use
                     (where x>=5 would mean only process samples with x>=5).
        :return: parameter standard deviation.
        """
        return np.sqrt(self.var(paramVec, where))

    def cov(self, pars=None, where=None):
        """
        Get parameter covariance

        :param pars: if specified, a list of parameter vectors or int indices to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: The covariance matrix
        """
        diffs = self.mean_diffs(pars, where)
        if pars is None:
            pars = list(range(self.n))
        n = len(pars)
        cov = np.empty((n, n))
        if where is not None:
            weights = self.weights[where]
        else:
            weights = self.weights
        for i, diff in enumerate(diffs):
            weightdiff = diff * weights
            for j in range(i, n):
                cov[i, j] = weightdiff.dot(diffs[j])
                cov[j, i] = cov[i, j]
        cov /= self.get_norm(where)
        return cov

    def corr(self, pars=None):
        """
        Get the correlation matrix

        :param pars: If specified, list of parameter vectors or int indices to use
        :return: The correlation matrix.
        """
        return covToCorr(self.cov(pars))

    def mean_diff(self, paramVec, where=None):
        """
        Calculates an array of differences between a parameter vector and the mean parameter value

        :param paramVec: array of parameter values or int index of parameter to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: array of p_i - mean(p_i)
        """
        if isinstance(paramVec, _int_types) and paramVec >= 0 and where is None:
            if self.diffs is not None:
                return self.diffs[paramVec]
            return self.samples[:, paramVec] - self.getMeans()[paramVec]
        paramVec = self._makeParamvec(paramVec)
        if where is None:
            return paramVec - self.mean(paramVec)
        else:
            return paramVec[where] - self.mean(paramVec, where)

    def mean_diffs(self, pars: Union[None, int, Sequence] = None, where=None) -> Sequence:
        """
        Calculates a list of parameter vectors giving distances from parameter means

        :param pars: if specified, list of parameter vectors or int parameter indices to use
        :param where: if specified, a filter for the samples to use
                      (where x>=5 would mean only process samples with x>=5).
        :return: list of arrays p_i-mean(p-i) for each parameter
        """
        if pars is None:
            pars = self.n
        if isinstance(pars, _int_types) and pars >= 0 and where is None:
            means = self.getMeans()
            return [self.samples[:, i] - means[i] for i in range(pars)]
        elif isinstance(pars, _int_types) and pars >= 0 and where is not None:
            return [self.mean_diff(i, where) for i in range(pars)]
        else:
            return [self.mean_diff(i, where) for i in pars]

    def twoTailLimits(self, paramVec, confidence):
        """
        Calculates two-tail equal-area confidence limit by counting samples in the tails

        :param paramVec: array of parameter values or int index of parameter to use
        :param confidence: confidence limit to calculate, e.g. 0.95 for 95% confidence
        :return: min, max values for the confidence interval
        """
        limits = np.array([(1 - confidence) / 2, 1 - (1 - confidence) / 2])
        return self.confidence(paramVec, limits)

    def initParamConfidenceData(self, paramVec, start=0, end=None, weights=None):
        """
        Initialize cache of data for calculating confidence intervals

        :param paramVec: array of parameter values or int index of parameter to use
        :param start: The sample start index to use
        :param end: The sample end index to use, use None to go all the way to the end of the vector
        :param weights: A numpy array of weights for each sample, defaults to self.weights
        :return: :class:`~.chains.ParamConfidenceData` instance
        """
        if weights is None:
            weights = self.weights
        paramVec = self._makeParamvec(paramVec)[start:end]
        indices = paramVec.argsort()
        return ParamConfidenceData(paramVec=paramVec,
                                   norm=np.sum(weights[start:end]),
                                   indexes=indices,
                                   cumsum=np.cumsum(weights[start + indices]))

    def confidence(self, paramVec, limfrac, upper=False, start=0, end=None, weights=None):
        """
        Calculate sample confidence limits, not using kernel densities just counting samples in the tails

        :param paramVec: array of parameter values or int index of parameter to use
        :param limfrac: fraction of samples in the tail,
                        e.g. 0.05 for a 95% one-tail limit, or 0.025 for a 95% two-tail limit
        :param upper: True to get upper limit, False for lower limit
        :param start: Start index for the vector to use
        :param end: The end index, use None to go all the way to the end of the vector.
        :param weights:  numpy array of weights for each sample, by default self.weights
        :return: confidence limit (parameter value when limfac of samples are further in the tail)
        """
        if isinstance(paramVec, ParamConfidenceData):
            d = paramVec
        else:
            d = self.initParamConfidenceData(paramVec, start, end, weights)

        if not upper:
            target = d.norm * limfrac
        else:
            target = d.norm * (1 - limfrac)
        ix = np.searchsorted(d.cumsum, target)
        return d.paramVec[d.indexes[np.minimum(ix, d.indexes.shape[0] - 1)]]

    def getSignalToNoise(self, params, noise=None, R=None, eigs_only=False):
        """
        Returns w, M, where w is the eigenvalues of the signal to noise (small y better constrained)

        :param params: list of parameters indices to use
        :param noise: noise matrix
        :param R: rotation matrix, defaults to inverse of Cholesky root of the noise matrix
        :param eigs_only: only return eigenvalues
        :return: w, M, where w is the eigenvalues of the signal to noise (small y better constrained)
        """
        C = self.cov(params)
        return getSignalToNoise(C, noise, R, eigs_only)

    def thin_indices(self, factor, weights=None):
        """
        Indices to make single weight 1 samples. Assumes integer weights.

        :param factor: The factor to thin by, should be int.
        :param weights: The weights to thin, None if this should use the weights stored in the object.
        :return: array of indices of samples to keep
        """
        if weights is None:
            weights = self.weights
        return WeightedSamples.thin_indices_single_samples(factor, weights)

    @staticmethod
    def thin_indices_and_weights(factor, weights):
        """
        Returns indices and new weights for use when thinning samples.

        :param factor: thin factor
        :param weights: initial weight (counts) per sample point
        :return: (unique index, counts) tuple of sample index values to keep and new weights
        """
        thin_ix = WeightedSamples.thin_indices_single_samples(factor, weights)
        return np.unique(thin_ix, return_counts=True)

    @staticmethod
    def thin_indices_single_samples(factor, weights):
        numrows = len(weights)
        norm1 = np.sum(weights)
        weights = weights.astype(int)
        norm = np.sum(weights)

        if abs(norm - norm1) > 1e-4:
            raise WeightedSampleError('Can only thin with integer weights')
        if factor != int(factor):
            raise WeightedSampleError('Thin factor must be integer')
        factor = int(factor)
        if factor >= np.max(weights):
            cumsum = np.cumsum(weights) // factor
            # noinspection PyTupleAssignmentBalance
            _, thin_ix = np.unique(cumsum, return_index=True)
        else:
            tot = 0
            i = 0
            thin_ix = np.empty(norm // factor, dtype=int)
            ix = 0
            mult = weights[i]
            while i < numrows:
                if mult + tot < factor:
                    tot += mult
                    i += 1
                    if i < numrows:
                        mult = weights[i]
                else:
                    thin_ix[ix] = i
                    ix += 1
                    if mult == factor - tot:
                        i += 1
                        if i < numrows:
                            mult = weights[i]
                    else:
                        mult -= (factor - tot)
                    tot = 0

        return thin_ix

    def random_single_samples_indices(self, random_state=None):
        """
        Returns an array of sample indices that give a list of weight-one samples, by randomly
        selecting samples depending on the sample weights

        :param random_state: random seed or Generator
        :return: array of sample indices
        """
        random_state = np.random.default_rng(random_state)
        max_weight = np.max(self.weights)
        thin_ix = []
        for i in range(self.numrows):
            P = self.weights[i] / max_weight
            if random_state.random() < P:
                thin_ix.append(i)
        return np.array(thin_ix, dtype=int)

    def thin(self, factor: int):
        """
        Thin the samples by the given factor, giving set of samples with unit weight

        :param factor: The factor to thin by
        """
        thin_ix = self.thin_indices(factor)
        self.setSamples(self.samples[thin_ix, :], loglikes=None if self.loglikes is None else self.loglikes[thin_ix],
                        min_weight_ratio=-1)

    def weighted_thin(self, factor: int):
        """
        Thin the samples by the given factor, preserving the weights (not all set to 1).

        :param factor: The (integer) factor to thin by
        """
        unique, counts = self.thin_indices_and_weights(factor, self.weights)
        self.setSamples(self.samples[unique, :],
                        loglikes=None if self.loglikes is None else self.loglikes[unique],
                        weights=counts, min_weight_ratio=-1)

    def filter(self, where):
        """
        Filter the stored samples to keep only samples matching filter

        :param where: list of sample indices to keep, or boolean array filter (e.g. x>5 to keep only samples where x>5)
        """
        self.setSamples(self.samples[where, :], self.weights[where],
                        None if self.loglikes is None else self.loglikes[where], min_weight_ratio=-1)

    def reweightAddingLogLikes(self, logLikes):
        """
        Importance sample the samples, by adding logLike (array of -log(likelihood values) to the currently
        stored likelihoods, and re-weighting accordingly, e.g. for adding a new data constraint

        :param logLikes: array of -log(likelihood) for each sample to adjust
        """
        scale = np.min(logLikes)
        if self.loglikes is not None:
            self.loglikes += logLikes
        self.weights = np.asarray(self.weights, dtype=np.float64)
        self.weights *= np.exp(-(logLikes - scale))
        self._weightsChanged()

    def cool(self, cool):
        """
        Cools the samples, i.e. multiples log likelihoods by cool factor and re-weights accordingly

        :param cool: cool factor
        """
        MaxL = np.max(self.loglikes)
        newL = self.loglikes * cool
        self.weights = np.asarray(self.weights, dtype=np.float64)
        self.weights *= np.exp(-(newL - self.loglikes) - (MaxL * (1 - cool)))
        self.loglikes = newL
        self._weightsChanged()

    def deleteZeros(self):
        """
        Removes samples with zero weight

        """
        self.filter(self.weights > 0)

    def setMinWeightRatio(self, min_weight_ratio=1e-30):
        """
        Removes samples with weight less than min_weight_ratio times the maximum weight

        :param min_weight_ratio: minimum ratio to max to exclude
        """
        if self.weights is not None and min_weight_ratio >= 0:
            max_weight = np.max(self.weights)
            min_weight = np.min(self.weights)
            if min_weight < max_weight * min_weight_ratio:
                self.filter(self.weights > max_weight * min_weight_ratio)

    def deleteFixedParams(self):
        """
        Removes parameters that do not vary (are the same in all samples)

        :return: tuple (list of fixed parameter indices that were removed, fixed values)
        """
        fixed = []
        values = []
        for i in range(self.samples.shape[1]):
            if np.isclose(self.samples[0, i], self.samples[-1, i]):
                mean = np.average(self.samples[:, i])
                if np.allclose(self.samples[:, i], mean, rtol=1e-12, atol=0):
                    fixed.append(i)
                    values.append(mean)
        self.changeSamples(np.delete(self.samples, fixed, 1))
        return fixed, values

    def removeBurn(self, remove=0.3):
        """
        removes burn in from the start of the samples

        :param remove: fraction of samples to remove, or if int >1, the number of sample rows to remove
        """
        if remove >= 1:
            ix = int(remove)
        else:
            ix = int(round(self.numrows * remove))
        if self.weights is not None:
            self.weights = self.weights[ix:]
        if self.loglikes is not None:
            self.loglikes = self.loglikes[ix:]
        self.changeSamples(self.samples[ix:, :])

    def saveAsText(self, root, chain_index=None, make_dirs=False):
        """
        Saves the samples as text files

        :param root: The root name to use
        :param chain_index: Optional index to be used for the samples' filename, zero based, e.g. for saving
                            one of multiple chains
        :param make_dirs: True if this should create the directories if necessary.
        """
        if self.loglikes is not None:
            loglikes = self.loglikes
        else:
            loglikes = np.zeros(self.numrows)
        if make_dirs and not os.path.exists(os.path.dirname(root)):
            os.makedirs(os.path.dirname(root))
        if root.endswith('.txt'):
            root = root[:-3]
        # noinspection PyTypeChecker
        np.savetxt(root + ('' if chain_index is None else '_' + str(chain_index + 1)) + '.txt',
                   np.hstack((self.weights.reshape(-1, 1), loglikes.reshape(-1, 1), self.samples)),
                   fmt=self.precision)

    def __getitem__(self, item):
        return self._makeParamvec(item)


# noinspection PyAttributeOutsideInit
class Chains(WeightedSamples):
    """
    Holds one or more sets of weighted samples, for example a set of MCMC chains.
    Inherits from :class:`~.chains.WeightedSamples`, also adding parameter names and labels

    :ivar paramNames: a :class:`~.paramnames.ParamNames` instance holding the parameter names and labels
    """

    def __init__(self, root=None, jobItem=None, paramNamesFile=None, names=None, labels=None, renames=None,
                 sampler=None, **kwargs):
        """

        :param root: optional root name for files
        :param jobItem: optional jobItem for parameter grid item. Should have jobItem.chainRoot and jobItem.batchPath
        :param paramNamesFile: optional filename of a .paramnames files that holds parameter names
        :param names: optional list of names for the parameters
        :param labels: optional list of latex labels for the parameters
        :param renames: optional dictionary of parameter aliases
        :param sampler: string describing the type of samples (default :mcmc); if "nested" or "uncorrelated"
              the effective number of samples is calculated using uncorrelated approximation
        :param kwargs: extra options for :class:`~.chains.WeightedSamples`'s constructor

        """

        self.chains = None
        self.chain_offsets = None
        super().__init__(**kwargs)
        self.jobItem = jobItem
        self.ignore_lines = float(kwargs.get('ignore_rows', 0))
        self.root = root
        if not paramNamesFile and root:
            if os.path.exists(root + '.paramnames'):
                paramNamesFile = root + '.paramnames'
            else:
                paramNamesFile = cobaya_interface.cobaya_params_file(root)
        self.setParamNames(paramNamesFile or names)
        if labels is not None:
            self.paramNames.setLabels(labels)
        if renames is not None:
            self.updateRenames(renames)
        # Sampler that generated the chain -- assume "mcmc"
        if isinstance(sampler, str):
            self.setSampler(sampler)
        else:
            self.sampler = "mcmc"

    def setSampler(self, sampler):
        sampler = sampler.lower()
        if sampler not in ["mcmc", "nested", "uncorrelated"]:
            raise ValueError("Unknown sampler type %s" % sampler)
        self.sampler = sampler

    def setParamNames(self, names=None):
        """
        Sets the names of the params.

        :param names: Either a :class:`~.paramnames.ParamNames` object, the name of a .paramnames file to load, a list
                      of name strings, otherwise use default names (param1, param2...).
        """
        self.paramNames = None
        if isinstance(names, ParamNames):
            self.paramNames = deepcopy(names)
        elif isinstance(names, str):
            self.paramNames = ParamNames(names)
        elif names is not None:
            self.paramNames = ParamNames(names=names)
        elif self.samples is not None:
            self.paramNames = ParamNames(default=self.n)
        if self.paramNames:
            self._getParamIndices()
        self.needs_update = True

    def filter(self, where):
        """
        Filter the stored samples to keep only samples matching filter

        :param where: list of sample indices to keep, or boolean array filter (e.g. x>5 to keep only samples where x>5)
        """

        if self.chains is None:
            if self.chain_offsets is not None:
                # must update chain_offsets to be able to correctly split back into separate filtered chains if needed
                lens = [0]
                for off1, off2 in zip(self.chain_offsets[:-1], self.chain_offsets[1:]):
                    lens.append(np.count_nonzero(where[off1:off2]))
                self.chain_offsets = np.cumsum(np.array(lens))
            super().filter(where)
        else:
            raise ValueError('chains are separated, makeSingle first or call filter on individual chains')

    def weighted_thin(self, factor: int):
        """
        Thin the samples by the given factor, giving (in general) non-unit integer weights.
        This function also preserves separate chains.

        :param factor: The (integer) factor to thin by
        """
        if not self.chains and self.chain_offsets is None:
            return super().weighted_thin(factor)
        has_chains = self.chains
        chains = self.getSeparateChains()
        for chain in chains:
            chain.weighted_thin(factor)
        self.chains = chains
        if not has_chains:
            self.makeSingle()
        self.needs_update = True

    def getParamNames(self):
        """
        Get :class:`~.paramnames.ParamNames` object with names for the parameters

        :return: :class:`~.paramnames.ParamNames` object giving parameter names and labels
        """
        return self.paramNames

    def _getParamIndices(self):
        """
        Gets the indices of the params.

        :return: A dict mapping the param name to the parameter index.
        """
        if self.samples is not None and len(self.paramNames.names) != self.n:
            raise WeightedSampleError("paramNames size (%s) does not match number of "
                                      "parameters in samples (%s)" % (len(self.paramNames.names), self.n))
        index = dict()
        for i, name in enumerate(self.paramNames.names):
            index[name.name] = i
        self.index = index
        return self.index

    def _parAndNumber(self, name):
        """
        Get index and ParamInfo for a name or index

        :param name: name or parameter index
        :return: index, ParamInfo instance
        """
        if isinstance(name, ParamInfo):
            name = name.name
        if isinstance(name, str):
            name = self.index.get(name, None)
            if name is None:
                return None, None
        if isinstance(name, _int_types):
            return name, self.paramNames.names[name]
        raise ParamError("Unknown parameter type %s" % name)

    def getRenames(self):
        """
        Gets dictionary of renames known to each parameter.
        """
        return self.paramNames.getRenames()

    def updateRenames(self, renames):
        """
        Updates the renames known to each parameter with the given dictionary of renames.
        """
        self.paramNames.updateRenames(renames)

    def setParams(self, obj):
        """
        Adds array variables obj.name1, obj.name2 etc, where
        obj.name1 is the vector of samples with name 'name1'

        if a parameter name is of the form aa.bb.cc, it makes subobjects so you can reference obj.aa.bb.cc.
        If aa.bb and aa are both parameter names, then aa becomes obj.aa.value.

        :param obj: The object instance to add the parameter vectors variables
        :return: The obj after alterations.
        """
        for second in [False, True]:
            for i, name in enumerate(self.paramNames.names):
                path = name.name.split('.')
                ob = obj
                for p in path[:-1]:
                    if not hasattr(ob, p):
                        setattr(ob, p, ParSamples())
                    ob = getattr(ob, p)
                if second:
                    if isinstance(getattr(ob, path[-1], None), ParSamples):
                        setattr(getattr(ob, path[-1]), 'value', self.samples[:, i])
                    else:
                        setattr(ob, path[-1], self.samples[:, i])

        return obj

    def getParams(self):
        """
        Creates a :class:`~.chains.ParSamples` object, with variables giving vectors for all the parameters,
        for example samples.getParams().name1 would be the vector of samples with name 'name1'

        :return: A :class:`~.chains.ParSamples` object containing all the parameter vectors, with attributes
                given by the parameter names
        """
        pars = ParSamples()
        self.setParams(pars)
        return pars

    def getParamSampleDict(self, ix, want_derived=True):
        """
        Returns a dictionary of parameter values for sample number ix

        :param ix: sample index
        :param want_derived: include derived parameters
        :return: ordered dictionary of parameter values
        """
        res = dict()
        res['weight'] = self.weights[ix]
        res['loglike'] = self.loglikes[ix]
        for i, name in enumerate(self.paramNames.names):
            if want_derived or not name.isDerived:
                res[name.name] = self.samples[ix, i]
        return res

    def _makeParamvec(self, par):
        if self.needs_update:
            self.updateBaseStatistics()
        if isinstance(par, ParamInfo):
            par = par.name
        if isinstance(par, str):
            index = self.index.get(par)
            if index is not None:
                return self.samples[:, index]
            if par == 'weight':
                return self.weights
            elif par == 'loglike':
                return self.loglikes
            else:
                raise ParamError('Unknown parameter %s' % par)
        return super()._makeParamvec(par)

    def updateChainBaseStatistics(self):
        # old name, use updateBaseStatistics
        return self.updateBaseStatistics()

    def updateBaseStatistics(self):
        """
        Updates basic computed statistics for this chain, e.g. after any changes to the samples or weights

        :return: self after updating statistics.
        """
        self.getVars()
        self.mean_mult = self.norm / self.numrows
        self.max_mult = np.max(self.weights)
        self._getParamIndices()
        self.needs_update = False
        return self

    def addDerived(self, paramVec, name, **kwargs):
        """
        Adds a new parameter

        :param paramVec: The vector of parameter values to add.
        :param name: The name for the new parameter
        :param kwargs: arguments for paramnames' :func:`.paramnames.ParamList.addDerived`
        :return: The added parameter's :class:`~.paramnames.ParamInfo` object
        """
        if self.paramNames.parWithName(name):
            raise ValueError('Parameter with name %s already exists' % name)
        self.changeSamples(np.c_[self.samples, paramVec])
        return self.paramNames.addDerived(name, **kwargs)

    def loadChains(self, root, files_or_samples: Sequence, weights=None, loglikes=None,
                   ignore_lines=None):
        """
        Loads chains from files.

        :param root: Root name
        :param files_or_samples: list of file names or list of arrays of samples, or single array of samples
        :param weights: if loading from arrays of samples, corresponding list of arrays of weights
        :param loglikes: if loading from arrays of samples, corresponding list of arrays of -2 log(likelihood)
        :param ignore_lines: Amount of lines at the start of the file to ignore, None if should not ignore
        :return: True if loaded successfully, False if none loaded
        """
        self.chains = []
        self.samples = None
        self.weights = None
        self.loglikes = None
        if ignore_lines is None:
            ignore_lines = self.ignore_lines
        WSkwargs = {"ignore_rows": ignore_lines,
                    "min_weight_ratio": self.min_weight_ratio}
        if hasattr(files_or_samples, '__len__') and not len(files_or_samples) or files_or_samples is None:
            raise ValueError('files_or_samples empty in loadChains')
        if isinstance(files_or_samples, str) or isinstance(files_or_samples[0], str):
            # From files
            if weights is not None or loglikes is not None:
                raise ValueError('weights and loglikes not needed reading from file')
            if isinstance(files_or_samples, str):
                files_or_samples = [files_or_samples]
            self.name_tag = self.name_tag or os.path.basename(root)
            for fname in files_or_samples:
                print_load_line(fname)
                try:
                    self.chains.append(WeightedSamples(fname, **WSkwargs))
                except WeightedSampleError:
                    print_load_line('Ignored file %s (likely empty)' % fname)
            nchains = len(self.chains)
            if not nchains:
                raise WeightedSampleError('loadChains - no chains found for ' + root)
        else:
            # From arrays
            def array_dimension(a):
                # Dimension for numpy or list/tuple arrays, not very safe (does not work if string elements)
                d = 0
                while True:
                    try:
                        a = a[0]
                        d += 1
                    except (TypeError, IndexError):
                        return d

            dim = array_dimension(files_or_samples)
            if dim in [1, 2]:
                self.chains = None
                self.setSamples(slice_or_none(files_or_samples, ignore_lines),
                                slice_or_none(weights, ignore_lines),
                                slice_or_none(loglikes, ignore_lines), self.min_weight_ratio)
                if self.paramNames is None:
                    self.paramNames = ParamNames(default=self.n)
                nchains = 1
            elif dim == 3:
                for i, samples_i in enumerate(files_or_samples):
                    self.chains.append(WeightedSamples(
                        samples=samples_i, loglikes=None if loglikes is None else loglikes[i],
                        weights=None if weights is None else weights[i], **WSkwargs))
                if self.paramNames is None:
                    self.paramNames = ParamNames(default=self.chains[0].n)
                nchains = len(self.chains)
            else:
                raise ValueError('samples or files must be array of samples, or a list of arrays or files')
        self._weightsChanged()
        return nchains > 0

    def getGelmanRubinEigenvalues(self, nparam=None, chainlist=None):
        """
        Assess convergence using var(mean)/mean(var) in the orthogonalized parameters
        c.f. Brooks and Gelman 1997.

        :param nparam: The number of parameters (starting at first), by default uses all of them
        :param chainlist: list of :class:`~.chains.WeightedSamples`, the samples to use.
                          Defaults to all the separate chains in this instance.
        :return: array of  var(mean)/mean(var) for orthogonalized parameters
        """
        if chainlist is None:
            chainlist = self.getSeparateChains()
        nparam = nparam or self.paramNames.numNonDerived()
        meanscov = np.zeros((nparam, nparam))
        means = self.getMeans()[:nparam]
        meancov = np.zeros(meanscov.shape)
        for chain in chainlist:
            diff = chain.getMeans()[:nparam] - means
            meanscov += np.outer(diff, diff)
            meancov += chain.getCov(nparam)
        meanscov /= (len(chainlist) - 1)
        meancov /= len(chainlist)
        w, U = np.linalg.eigh(meancov)
        if np.min(w) > 0:
            U /= np.sqrt(w)
            D = np.linalg.eigvalsh(np.dot(U.T, meanscov).dot(U))
            return D
        else:
            return None

    def getGelmanRubin(self, nparam=None, chainlist=None):
        """
        Assess the convergence using the maximum var(mean)/mean(var) of orthogonalized parameters
        c.f. Brooks and Gelman 1997.

        :param nparam: The number of parameters, by default uses all
        :param chainlist: list of :class:`~.chains.WeightedSamples`, the samples to use. Defaults to all the
                          separate chains in this instance.
        :return: The worst var(mean)/mean(var) for orthogonalized parameters. Should be <<1 for good convergence.
        """
        return np.max(self.getGelmanRubinEigenvalues(nparam, chainlist))

    def makeSingle(self):
        """
        Combines separate chains into one samples array, so self.samples has all the samples
        and this instance can then be used as a general :class:`~.chains.WeightedSamples` instance.

        :return: self
        """
        if not self.chains:
            raise ValueError('There are no separated chains for makeSingle()')
        self.chain_offsets = np.cumsum(np.array([0] + [chain.samples.shape[0] for chain in self.chains]))
        weights = None if self.chains[0].weights is None else np.hstack([chain.weights for chain in self.chains])
        loglikes = None if self.chains[0].loglikes is None else np.hstack([chain.loglikes for chain in self.chains])
        self.setSamples(np.vstack([chain.samples for chain in self.chains]), weights, loglikes, min_weight_ratio=-1)
        self.chains = None
        self.needs_update = True
        return self

    def getSeparateChains(self) -> List['WeightedSamples']:
        """
        Gets a list of samples for separate chains.
        If the chains have already been combined, uses the stored sample offsets to reconstruct the array
        (generally no array copying)

        :return: The list of :class:`~.chains.WeightedSamples` for each chain.
        """
        if self.chains is not None:
            return self.chains
        chainlist = []
        if self.chain_offsets is None:
            raise WeightedSampleError('Samples were not combined from separate chains')
        else:
            for off1, off2 in zip(self.chain_offsets[:-1], self.chain_offsets[1:]):
                chainlist.append(WeightedSamples(samples=self.samples[off1:off2], weights=self.weights[off1:off2],
                                                 loglikes=self.loglikes[off1:off2]))
        return chainlist

    def removeBurnFraction(self, ignore_frac):
        """
        Remove a fraction of the samples as burn in

        :param ignore_frac: fraction of sample points to remove from the start of the samples, or each chain
                            if not combined
        """
        if self.samples is not None:
            self.removeBurn(ignore_frac)
            self.chains = None
            self.needs_update = True
        else:
            for chain in self.chains:
                chain.removeBurn(ignore_frac)

    def deleteFixedParams(self):
        """
        Delete parameters that are fixed (the same value in all samples)
        """
        if self.samples is not None:
            fixed, values = super().deleteFixedParams()
            self.chains = None
        else:
            fixed, values = self.chains[0].deleteFixedParams()
            for chain in self.chains[1:]:
                chain.changeSamples(np.delete(chain.samples, fixed, 1))
        if hasattr(self, 'ranges'):
            for ix, value in zip(fixed, values):
                self.ranges.setFixed(self.paramNames.names[ix].name, value)
        self.paramNames.deleteIndices(fixed)
        self._getParamIndices()

    def saveAsText(self, root, chain_index=None, make_dirs=False):
        """
        Saves the samples as text files, including parameter names as .paramnames file.

        :param root: The root name to use
        :param chain_index: Optional index to be used for the filename, zero based, e.g. for saving one
                            of multiple chains
        :param make_dirs: True if this should (recursively) create the directory if it doesn't exist
        """
        super().saveAsText(root, chain_index, make_dirs)
        if not chain_index:
            self.saveTextMetadata(root)

    def saveTextMetadata(self, root):
        """
        Saves metadata about the sames to text files with given file root

        :param root: root file name
        """
        self.paramNames.saveAsText(root + '.paramnames')

    def savePickle(self, filename):
        """
        Save the current object to a file in pickle format

        :param filename: The file to write to
        """

        with open(filename, 'wb') as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)



# usage:
# python covcmb.py out.covmat in1.covmat in2.covmat
# Nb. in1 values take priority over in2

import sys
from getdist import covmat

if len(sys.argv) < 3:
    print('Usage: python covcmb.py out.covmat in1.covmat in2.covmat [in3.covmat...]')
    sys.exit()

foutname = sys.argv[1]

cov = covmat.CovMat(sys.argv[2])

for fname in sys.argv[3:]:
    print('merging: ' + fname)
    cov = cov.mergeCovmatWhereNew(covmat.CovMat(fname))

cov.saveToFile(foutname)



import tempfile
import os
import numpy as np
import unittest
import subprocess
import shutil
from getdist import loadMCSamples, plots, IniFile
from getdist.tests.test_distributions import Test2DDistributions, Gaussian1D, Gaussian2D
from getdist.mcsamples import MCSamples
from getdist.styles.tab10 import style_name as tab10
from getdist.styles.planck import style_name as planck
from getdist.parampriors import ParamBounds
from matplotlib import rcParams
import matplotlib.pyplot as plt


class GetDistFileTest(unittest.TestCase):
    """test reading files, convergence routines and getdist script"""

    def setUp(self):
        random_state = np.random.default_rng(10)
        # Simulate some chain files
        prob = Test2DDistributions().bimodal[0]
        self.tempdir = os.path.join(tempfile.gettempdir(), 'gettdist_tests')
        if not os.path.exists(self.tempdir):
            os.mkdir(self.tempdir)
        self.root = os.path.join(self.tempdir, 'testchain')
        for n in range(3):
            mcsamples = prob.MCSamples(4000, logLikes=True, random_state=random_state)
            mcsamples.saveAsText(self.root, chain_index=n)

    def tearDown(self):
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.tempdir)

    def testFileLoadPlot(self):
        samples = loadMCSamples(self.root, settings={'ignore_rows': 0.1})
        g = plots.get_single_plotter(chain_dir=self.tempdir, analysis_settings={'ignore_rows': 0.1})
        self.assertEqual(g.samples_for_root('testchain').numrows, samples.numrows,
                         "Inconsistent chain loading")
        self.assertEqual(g.samples_for_root('testchain').getTable().tableTex(),
                         samples.getTable().tableTex(), 'Inconsistent load result')
        samples.getConvergeTests(0.95)
        self.assertAlmostEqual(0.00052997, samples.GelmanRubin, 4, 'Gelman Rubin error, got ' + str(samples.GelmanRubin))

        g = plots.get_single_plotter()
        g.plot_3d(samples, ['x', 'y', 'x'])
        g.export(self.root + '_plot.pdf')

        g = plots.get_single_plotter(chain_dir=self.tempdir,
                                     analysis_settings={'ignore_rows': 0.1, 'contours': [0.68, 0.95, 0.99]})
        g.settings.num_plot_contours = 3
        g.plot_2d('testchain', ['x', 'y'])

    def testGetDist(self):
        from getdist.command_line import getdist_command

        os.chdir(self.tempdir)
        res = getdist_command([self.root])
        # Note this can fail if your local analysis defaults changes the default ignore_rows
        self.assertTrue('-Ln(mean like)  = 2.31' in res, res)
        fname = 'testchain_pars.ini'
        getdist_command(['--make_param_file', fname])
        ini = IniFile(fname)
        ini.params['no_plots'] = False
        ini.params['plot_2D_num'] = 1
        ini.params['plot1'] = 'x y'
        ini.params['num_3D_plots'] = 1
        ini.params['3D_plot1'] = 'x y x'
        ini.params['triangle_params'] = '*[xy]*'

        ini.saveFile(fname)
        res = getdist_command([fname, self.root])
        self.assertTrue('-Ln(mean like)  = 2.31' in res)

        def check_run():
            for f in ['.py', '_2D.py', '_3D.py', '_tri.py']:
                pyname = self.root + f
                self.assertTrue(os.path.isfile(pyname))
                subprocess.check_output(['python', pyname])
                pdf = self.root + f.replace('py', 'pdf')
                self.assertTrue(os.path.isfile(pdf))
                os.remove(pdf)
                os.remove(pyname)

        check_run()


class GetDistTest(unittest.TestCase):
    """test some getdist routines and plotting"""

    def setUp(self):
        self.testdists = Test2DDistributions()

    def testBestFit(self):
        samples = self.testdists.bimodal[0].MCSamples(12000, logLikes=True, random_state=10)
        bestSample = samples.getParamBestFitDict(best_sample=True)
        self.assertAlmostEqual(bestSample['loglike'], 1.708, 2)

    def testTables(self):
        self.samples = self.testdists.bimodal[0].MCSamples(12000, logLikes=True, random_state=10)
        self.assertEqual(str(self.samples.getLatex(limit=2)),
                         "(['x', 'y'], ['0.0^{+2.1}_{-2.1}', '0.0^{+1.3}_{-1.3}'])", "MCSamples.getLatex error")
        table = self.samples.getTable(columns=1, limit=1, paramList=['x'])
        self.assertTrue(r'0.0\pm 1.2' in table.tableTex(), "Table tex error")

    def testPCA(self):
        samples = self.testdists.bending.MCSamples(12000, logLikes=True, random_state=10)
        self.assertTrue('e-value: 0.10' in samples.PCA(['x', 'y']))

    def testLimits(self):
        samples = self.testdists.cut_correlated.MCSamples(12000, logLikes=False, random_state=10)
        stats = samples.getMargeStats()
        lims = stats.parWithName('x').limits
        self.assertAlmostEqual(lims[0].lower, 0.2077, 3)
        self.assertAlmostEqual(lims[1].lower, 0.0574, 3)
        self.assertTrue(lims[2].onetail_lower)

        # check some analytics (note not very accurate actually)
        samples = Gaussian1D(0, 1, xmax=1).MCSamples(1500000, logLikes=False, random_state=10)
        stats = samples.getMargeStats()
        lims = stats.parWithName('x').limits
        self.assertAlmostEqual(lims[0].lower, -0.78828, 2)
        self.assertAlmostEqual(lims[0].upper, 0.7954, 2)
        self.assertAlmostEqual(lims[1].lower, -1.730, 2)

    def testDensitySymmetries(self):
        # check flipping samples gives flipped density
        samps = Gaussian1D(0, 1, xmin=-1, xmax=3).MCSamples(12000, random_state=10)
        d = samps.get1DDensity('x')
        samps.samples[:, 0] *= -1
        samps = MCSamples(samples=samps.samples, names=['x'], ranges={'x': [-3, 1]})
        d2 = samps.get1DDensity('x')
        self.assertTrue(np.allclose(d.P, d2.P[::-1]))

        samps = Gaussian2D([0, 0], np.diagflat([1, 2]), xmin=-1, xmax=2, ymin=0, ymax=3).MCSamples(12000,
                                                                                                   random_state=10)
        d = samps.get2DDensity('x', 'y')
        samps.samples[:, 0] *= -1
        samps = MCSamples(samples=samps.samples, names=['x', 'y'], ranges={'x': [-2, 1], 'y': [0, 3]})
        d2 = samps.get2DDensity('x', 'y')
        self.assertTrue(np.allclose(d.P, d2.P[:, ::-1]))
        samps.samples[:, 0] *= -1
        samps.samples[:, 1] *= -1
        samps = MCSamples(samples=samps.samples, names=['x', 'y'], ranges={'x': [-1, 2], 'y': [-3, 0]})
        d2 = samps.get2DDensity('x', 'y')
        self.assertTrue(np.allclose(d.P, d2.P[::-1, ::], atol=1e-5))

    def testLoads(self):
        # test initiating from multiple chain arrays
        samps = []
        for i in range(3):
            samps.append(Gaussian2D([1.5, -2], np.diagflat([1, 2])).MCSamples(1001 + i * 10, names=['x', 'y'],
                                                                              random_state=10))
        fromChains = MCSamples(samples=[s.samples for s in samps], names=['x', 'y'])
        mean = np.sum([s.norm * s.mean('x') for s in samps]) / np.sum([s.norm for s in samps])
        meanChains = fromChains.mean('x')
        self.assertAlmostEqual(mean, meanChains)
        self.assertAlmostEqual(mean, float(np.mean(fromChains['x'])))

    def testMixtures(self):
        from getdist.gaussian_mixtures import Mixture2D, GaussianND

        cov1 = [[0.001 ** 2, 0.0006 * 0.05], [0.0006 * 0.05, 0.05 ** 2]]
        cov2 = [[0.01 ** 2, -0.005 * 0.03], [-0.005 * 0.03, 0.03 ** 2]]
        mean1 = [0.02, 0.2]
        mean2 = [0.023, 0.09]
        mixture = Mixture2D([mean1, mean2], [cov1, cov2], names=['zobs', 't'], labels=[r'z_{\rm obs}', 't'],
                            label='Model')
        tester = 0.03
        cond = mixture.conditionalMixture(['zobs'], [tester])
        marge = mixture.marginalizedMixture(['zobs'])
        # test P(x,y) = P(y)P(x|y)
        self.assertAlmostEqual(mixture.pdf([tester, 0.15]), marge.pdf([tester]) * cond.pdf([0.15]))

        samples = mixture.MCSamples(3000, label='Samples', random_state=10)
        g = plots.get_subplot_plotter()
        g.triangle_plot([samples, mixture], filled=False)
        g.new_plot()
        g.plot_1d(cond, 't')

        s1 = 0.0003
        covariance = [[s1 ** 2, 0.6 * s1 * 0.05, 0], [0.6 * s1 * 0.05, 0.05 ** 2, 0.2 ** 2], [0, 0.2 ** 2, 2 ** 2]]
        mean = [0.017, 1, -2]
        gauss = GaussianND(mean, covariance)
        g = plots.get_subplot_plotter()
        g.triangle_plot(gauss, filled=True)

    def testPlots(self):
        self.samples = self.testdists.bimodal[0].MCSamples(12000, logLikes=True, random_state=10)
        g = plots.get_single_plotter(auto_close=True)
        samples = self.samples
        p = samples.getParams()
        samples.addDerived(p.x + (5 + p.y) ** 2, name='z')
        samples.addDerived(p.x, name='x.yx', label='forPattern')
        samples.addDerived(p.y, name='x.2', label='x_2')
        samples.updateBaseStatistics()

        g.plot_1d(samples, 'x')
        g.new_plot()
        g.plot_1d(samples, 'y', normalized=True, marker=0.1, marker_color='b')
        g.new_plot()
        g.plot_2d(samples, 'x', 'y')
        g.new_plot()
        g.plot_2d(samples, 'x', 'y', filled=True)
        g.new_plot()
        g.plot_2d(samples, 'x', 'y', shaded=True)
        g.new_plot()
        g.plot_2d_scatter(samples, 'x', 'y', color='red', colors=['blue'])
        g.new_plot()
        g.plot_3d(samples, ['x', 'y', 'z'])

        g = plots.get_subplot_plotter(width_inch=8.5, auto_close=True)
        g.plots_1d(samples, ['x', 'y'], share_y=True)
        g.new_plot()
        g.triangle_plot(samples, ['x', 'y', 'z'])
        self.assertTrue(g.get_axes_for_params('x', 'z') == g.subplots[2, 0])
        self.assertTrue(g.get_axes_for_params('z', 'x', ordered=False) == g.subplots[2, 0])
        self.assertTrue(g.get_axes_for_params('x') == g.subplots[0, 0])
        self.assertTrue(g.get_axes_for_params('x', 'p', 'q') is None)
        self.assertTrue(g.get_axes(ax=('x', 'z')) == g.subplots[2, 0])
        self.assertTrue(g.get_axes(ax=(2, 0)) == g.subplots[2, 0])

        g.new_plot()
        g.triangle_plot(samples, ['x', 'y'], plot_3d_with_param='z')
        g.new_plot()
        g.rectangle_plot(['x', 'y'], ['z'], roots=samples, filled=True)
        prob2 = self.testdists.bimodal[1]
        samples2 = prob2.MCSamples(12000, random_state=10)
        g.new_plot()
        g.triangle_plot([samples, samples2], ['x', 'y'])
        g.new_plot()
        g.plots_2d([samples, samples2], param_pairs=[['x', 'y'], ['x', 'z']])
        g.new_plot()
        g.plots_2d([samples, samples2], 'x', ['z', 'y'])
        g.new_plot()
        self.assertEqual([name.name for name in samples.paramNames.parsWithNames('x.*')], ['x.yx', 'x.2'])
        g.triangle_plot(samples, 'x.*')
        samples.updateSettings({'contours': '0.68 0.95 0.99'})
        g.settings.num_plot_contours = 3
        g.plot_2d(samples, 'x', 'y', filled=True)
        g.add_y_bands(0.2, 1.5)
        g.add_x_bands(-0.1, 1.2, color='red')
        g.new_plot()
        omm = np.arange(0.1, 0.7, 0.01)
        g.add_bands(omm, 0.589 * omm ** (-0.25), 0.03 * omm ** (-0.25), nbands=3)

        g = plots.get_subplot_plotter()
        import copy
        for upper in [False, True]:
            g.triangle_plot([samples, samples2], ['x', 'y', 'z'], filled=True,
                            upper_roots=[copy.deepcopy(samples)], upper_kwargs={'contour_colors': ['green']},
                            legend_labels=['1', '2', '3'], upper_label_right=upper)
            for i in range(3):
                for j in range(i):
                    self.assertTrue(g.subplots[i, j].get_xlim() == g.subplots[j, i].get_ylim())
                    self.assertTrue(g.subplots[i, j].get_ylim() == g.subplots[j, i].get_xlim())
                    self.assertTrue(g.subplots[i, j].get_xlim() == g.subplots[j, j].get_xlim())

    def test_styles(self):
        tmp = rcParams.copy()
        plots.set_active_style(tab10)
        g = plots.get_single_plotter()
        self.assertEqual(g.settings.line_styles.name, 'tab10')
        plots.set_active_style(planck)
        g = plots.get_single_plotter()
        self.assertTrue(g.settings.prob_y_ticks)
        plots.set_active_style(tab10)
        g = plots.get_single_plotter()
        self.assertEqual(g.settings.line_styles.name, 'tab10')
        plots.set_active_style()
        g = plots.get_single_plotter()
        self.assertFalse(g.settings.prob_y_ticks)
        g = plots.get_single_plotter(style='tab10')
        self.assertEqual(g.settings.line_styles.name, 'tab10')
        plots.set_active_style('planck')
        plots.set_active_style()
        self.assertDictEqual(tmp, rcParams)


class UtilTest(unittest.TestCase):
    """test bounded and unbounded tick assignment"""

    def _plot_with_params(self, scale, x, off, prune, default=False):
        from getdist.matplotlib_ext import BoundedMaxNLocator

        fig, axs = plt.subplots(1, 1, figsize=(x, 1))
        axs.plot([off - scale, off + scale], [0, 1])
        axs.set_yticks([])
        if not default:
            axs.xaxis.set_major_locator(BoundedMaxNLocator(prune=prune))
        axs.xaxis.get_major_formatter().useOffset = False
        fig.suptitle("%s: scale %g, size %g, offset %g" % ('Default' if default else 'Bounded', scale, x, off),
                     fontsize=6)
        return fig, axs

    def test_one_locator(self):
        self._plot_with_params(0.01, 1, 0.05, True)
        plt.draw()

    def test_y(self):
        from getdist.matplotlib_ext import BoundedMaxNLocator
        fig, ax = plt.subplots(1, 1, figsize=(3, 3))
        ax.plot([0, 1], [0, 1])
        ax.yaxis.set_major_locator(BoundedMaxNLocator(prune=True))

        def check_ticks(bounds, expected):
            ax.set_ylim(bounds)
            ticks = ax.get_yticks()
            if len(ticks) != len(expected) or not np.allclose(ticks, expected):
                raise self.failureException("Wrong ticks %s for bounds %s" % (ticks, bounds))

        check_ticks([0.0253, 0.02915], [0.026, 0.027, 0.028])

    def test_specifics(self):
        testdists = Test2DDistributions()
        samples = testdists.bimodal[0].MCSamples(1000, logLikes=True, random_state=10)
        g = plots.get_subplot_plotter(auto_close=True)
        g.settings.prob_label = r'$P$'
        g.settings.prob_y_ticks = True
        g.plot_1d(samples, 'x', _no_finish=True)
        ax = g.get_axes()
        self.assertTrue(np.allclose(ax.get_yticks(), [0, 0.5, 1]), "Wrong probability ticks")

        def check_ticks(bounds, expected):
            ax.set_xlim(bounds)
            ticks = ax.get_xticks()
            if len(ticks) != len(expected) or not np.allclose(ticks, expected):
                raise self.failureException("Wrong ticks %s for bounds %s" % (ticks, bounds))

        check_ticks([-5.2, 5.2], [-4, -2, 0, 2, 4])
        check_ticks([0, 8.2], [0, 2, 4, 6, 8])
        check_ticks([0.0219, 0.02232], [0.022, 0.0222])
        check_ticks([-0.009, 0.009], [-0.008, 0., 0.008])
        g.make_figure(nx=2, ny=1, sharey=True)
        ax = g.get_axes()
        g._set_main_axis_properties(ax.xaxis, True)
        ax.set_yticks([])
        check_ticks([-0.009, 0.009], [-0.006, 0., 0.006])
        check_ticks([1, 1.0004], [1.0001, 1.0003])

    def test_locator(self):
        import matplotlib.backends.backend_pdf
        # Set TMPSMALL env variable to save the output PDF for inspection
        local = os.environ.get('TMPSMALL')
        temp = os.path.join(local or tempfile.gettempdir(), 'output.pdf')
        pdf = matplotlib.backends.backend_pdf.PdfPages(temp)
        fails = []
        for x in np.arange(1, 5, 0.5):
            for scale in [1e-4, 0.9e-2, 1e-1, 1, 14, 3000]:
                for off in [scale / 3, 1, 7.4 * scale]:
                    for prune in [True, False]:
                        fig, ax = self._plot_with_params(scale, x, off, prune)
                        pdf.savefig(fig, bbox_inches='tight')
                        if not len(ax.get_xticks()) or x >= 2 > len(ax.get_xticks()) and scale > 1e-4:
                            fails.append([scale, x, off, prune])
                        plt.close(fig)
                    if local:
                        fig, ax = self._plot_with_params(scale, x, off, True, True)
                        pdf.savefig(fig, bbox_inches='tight')
                        plt.close(fig)
        pdf.close()
        if not local:
            os.remove(temp)

        self.assertFalse(len(fails), "Too few ticks for %s" % fails)


class CobayaTest(unittest.TestCase):
    def setUp(self):
        self.tempdir = os.path.join(tempfile.gettempdir(), 'gettdist_tests')
        if not os.path.exists(self.tempdir):
            os.mkdir(self.tempdir)
        os.chdir(self.tempdir)
        self.path = os.getenv('TRAVIS_BUILD_DIR', os.path.join(os.path.dirname(__file__), '..', '..', '..'))
        self.path = os.path.normpath(os.path.join(self.path, 'getdist_testchains', 'cobaya'))

    def tearDown(self):
        os.chdir(tempfile.gettempdir())
        shutil.rmtree(self.tempdir)

    def test_chains(self):
        if os.path.exists(self.path):
            root = os.path.join(self.path, 'DES_shear')
            samples = loadMCSamples(root, settings={'ignore_rows': 0.3}, no_cache=True)
            self.assertAlmostEqual(samples.mean('ombh2'), 0.02764592190482377, 6)
            pars = samples.getParamSampleDict(10)
            self.assertAlmostEqual(0.06, pars['mnu'], 6)
            self.assertAlmostEqual(samples.getUpper('ns'), 1.07, 6)
            self.assertAlmostEqual(samples.getLower('ns'), 0.87, 6)
            self.assertEqual(samples.getLower('DES_DzS2'), None)
            self.assertAlmostEqual(0, pars['omk'])
            from getdist.command_line import getdist_command
            res = getdist_command([root])
            self.assertTrue('-log(Like) = 95.49' in res, res)

    def test_planck_chains(self):
        if os.path.exists(self.path):
            root = os.path.join(self.path, 'compare_devel_drag')
            samples = loadMCSamples(root, settings={'ignore_rows': 0.3}, no_cache=True)
            self.assertAlmostEqual(samples.mean('ombh2'), 0.0223749, 6)
            self.assertAlmostEqual(samples.getUpper('H0'), 100, 6)
            self.assertEqual(samples.getLower('sigma8'), None)
            samples.saveAsText(r'planck_test')
            ranges = ParamBounds('planck_test.ranges')
            for par in samples.paramNames.names:
                self.assertEqual(samples.getUpper(par.name), ranges.getUpper(par.name))



import os
import subprocess
import getdist
import sys
import logging
from getdist import MCSamples, chains, IniFile


def runScript(fname):
    subprocess.Popen(['python', fname])


# noinspection PyUnboundLocalVariable,PyProtectedMember
def getdist_script(args, exit_on_error=True):
    def do_error(msg):
        if exit_on_error:
            print(msg)
            sys.exit()
        raise ValueError(msg)

    result = []

    def doprint(*s):
        result.append(" ".join([str(x) for x in s]))
        print(*s)

    no_plots = False
    chain_root = args.chain_root
    if args.ini_file is None and chain_root is None:
        do_error('Must give either a .ini file of parameters or a chain file root name. Run "getdist -h" for help.')
    if '.ini' not in args.ini_file and chain_root is None:
        # use default settings acting on chain_root, no plots
        chain_root = args.ini_file
        args.ini_file = getdist.default_getdist_settings
        no_plots = True
    if not os.path.isfile(args.ini_file):
        do_error('Parameter file does not exist: ' + args.ini_file)
    if chain_root and chain_root.endswith('.txt'):
        chain_root = chain_root[:-4]

    if chain_root is not None and ('*' in chain_root or '?' in chain_root):
        import glob
        import copy
        for ending in ['.paramnames', 'updated.yaml']:
            for f in glob.glob(chain_root + ending):
                fileargs = copy.copy(args)
                fileargs.chain_root = f.replace(ending, '')
                getdist_script(fileargs)
        return

    # Input parameters
    ini = IniFile(args.ini_file)

    for item in set(ini.params.keys()).intersection(
            {'make_single_samples', 'single_thin', 'dump_ND_bins', 'plot_meanlikes', 'shade_meanlikes',
             'plot_data_dir', 'force_twotail'}):
        if ini.string(item) not in [0, 'F']:
            logging.warning('%s is no longer supported by getdist, value ignored' % item)

    # File root
    if chain_root is not None:
        in_root = chain_root
    else:
        in_root = ini.params['file_root']
    if not in_root:
        do_error('Chain Root file name not given ')
    rootname = os.path.basename(in_root)

    if args.ignore_rows is not None:
        ignorerows = args.ignore_rows
    else:
        ignorerows = ini.float('ignore_rows', 0.0)

    samples_are_chains = ini.bool('samples_are_chains', True)

    paramnames = ini.string('parameter_names', '')

    # Create instance of MCSamples
    mc = MCSamples(in_root, ini=ini, files_are_chains=samples_are_chains, paramNamesFile=paramnames)

    if ini.bool('adjust_priors', False) or ini.bool('map_params', False):
        do_error('To adjust priors or define new parameters, use a separate python script; '
                 'see the python getdist docs for examples')

    plot_ext = ini.string('plot_ext', 'py')
    finish_run_command = ini.string('finish_run_command', '')

    no_plots = ini.bool('no_plots', no_plots)
    plots_only = ini.bool('plots_only', False)
    no_tests = plots_only or ini.bool('no_tests', False)

    thin_factor = ini.int('thin_factor', 0)
    thin_cool = ini.float('thin_cool', 1.0)

    cool = ini.float('cool', 1.0)

    chain_exclude = ini.int_list('exclude_chain')

    out_dir = ini.string('out_dir', './')
    if out_dir:
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)
        doprint('producing files in directory ', out_dir)
    mc.out_dir = out_dir

    out_root = ini.string('out_root', '')
    if out_root:
        rootname = out_root
        doprint('producing files with with root ', out_root)
    mc.rootname = rootname

    rootdirname = os.path.join(out_dir, rootname)
    mc.rootdirname = rootdirname

    if 'do_minimal_1d_intervals' in ini.params:
        do_error('do_minimal_1d_intervals no longer used; set credible_interval_threshold instead')

    line = ini.string('PCA_params', '')
    if line.lower() == 'all':
        PCA_params = mc.paramNames.list()
    else:
        PCA_params = line.split()
    PCA_num = ini.int('PCA_num', len(PCA_params))
    if PCA_num != 0:
        if PCA_num < 2:
            do_error('Can only do PCA for 2 or more parameters')
        PCA_func = ini.string('PCA_func', '')
        # Characters representing functional mapping
        if PCA_func == '':
            PCA_func = ['N'] * PCA_num  # No mapping
        PCA_NormParam = ini.string('PCA_normparam', '') or None

    # ==============================================================================

    first_chain = ini.int('first_chain', 0)
    last_chain = ini.int('chain_num', -1)
    # -1 y keep reading until one not found

    # Chain files
    for separator in ['_', '.']:
        chain_files = chains.chainFiles(in_root, first_chain=first_chain, last_chain=last_chain,
                                        chain_exclude=chain_exclude, separator=separator)
        if chain_files:
            break

    mc.loadChains(in_root, chain_files)

    mc.removeBurnFraction(ignorerows)
    if chains.print_load_details:
        if ignorerows:
            doprint('Removed %s as burn in' % ignorerows)
        else:
            doprint('Removed no burn in')

    mc.deleteFixedParams()
    mc.makeSingle()

    def filterParList(namestring, num=None):
        if not namestring.strip():
            _pars = mc.paramNames.list()
        else:
            _pars = []
            for name in namestring.split():
                if '?' in name or '*' in name:
                    _pars += mc.paramNames.getMatches(name, strings=True)
                elif mc.paramNames.parWithName(name):
                    _pars.append(name)
        if num is not None and len(_pars) != num:
            doprint('%iD plot has missing parameter or wrong number of parameters: %s' % (num, _pars))
            _pars = None
        return _pars

    if cool != 1:
        doprint('Cooling chains by ', cool)
        mc.cool(cool)

    mc.updateBaseStatistics()

    if not no_tests:
        mc.getConvergeTests(mc.converge_test_limit, writeDataToFile=True, feedback=True)

    mc.writeCovMatrix()
    mc.writeCorrelationMatrix()

    # Output thinned data if requested
    # Must do this with unsorted output
    if thin_factor > 1:
        thin_ix = mc.thin_indices(thin_factor)
        filename = rootdirname + '_thin.txt'
        mc.writeThinData(filename, thin_ix, thin_cool)

    doprint(mc.getNumSampleSummaryText().strip())
    if mc.likeStats:
        doprint(mc.likeStats.likeSummary().strip())

    if PCA_num > 0 and not plots_only:
        mc.PCA(PCA_params, PCA_func, PCA_NormParam, writeDataToFile=True)

    # Do 1D bins
    mc._setDensitiesandMarge1D()

    if not no_plots:
        # Output files for 1D plots

        plotparams = []
        line = ini.string('plot_params', '')
        if line not in ['', '0']:
            plotparams = filterParList(line)

        line = ini.string('plot_2D_param', '').strip()
        plot_2D_param = None
        if line and line != '0':
            plot_2D_param = line

        cust2DPlots = []
        if not plot_2D_param:
            # Use custom array of specific plots
            num_cust2D_plots = ini.int('plot_2D_num', 0)
            for i in range(1, num_cust2D_plots + 1):
                line = ini.string('plot' + str(i))
                pars = filterParList(line, 2)
                if pars is not None:
                    cust2DPlots.append(pars)
                else:
                    num_cust2D_plots -= 1

        triangle_params = []
        triangle_plot = ini.bool('triangle_plot', False)
        if triangle_plot:
            line = ini.string('triangle_params', '')
            triangle_params = filterParList(line)
            triangle_num = len(triangle_params)
            triangle_plot = triangle_num > 1

        num_3D_plots = ini.int('num_3D_plots', 0)
        plot_3D = []
        for ix in range(1, num_3D_plots + 1):
            line = ini.string('3D_plot' + str(ix))
            pars = filterParList(line, 3)
            if pars is not None:
                plot_3D.append(pars)
            else:
                num_3D_plots -= 1

        make_plots = ini.bool('make_plots', False) or args.make_plots

        filename = rootdirname + '.' + plot_ext
        mc._writeScriptPlots1D(filename, plotparams)
        if make_plots:
            runScript(filename)

        # Do 2D bins
        if plot_2D_param == 'corr':
            # In this case output the most correlated variable combinations
            doprint('...doing 2D plots for most correlated variables')
            cust2DPlots = mc.getCorrelatedVariable2DPlots()
            plot_2D_param = None
        elif plot_2D_param:
            mc.paramNames.parWithName(plot_2D_param, error=True)  # just check

        if cust2DPlots or plot_2D_param:
            doprint('...producing 2D plots')
            filename = rootdirname + '_2D.' + plot_ext
            mc._writeScriptPlots2D(filename, plot_2D_param, cust2DPlots)
            if make_plots:
                runScript(filename)

        if triangle_plot:
            # Add the off-diagonal 2D plots
            doprint('...producing triangle plot')
            filename = rootdirname + '_tri.' + plot_ext
            mc._writeScriptPlotsTri(filename, triangle_params)
            if make_plots:
                runScript(filename)

        # Do 3D plots (i.e. 2D scatter plots with coloured points)
        if num_3D_plots:
            doprint('...producing ', num_3D_plots, '2D colored scatter plots')
            filename = rootdirname + '_3D.' + plot_ext
            mc._writeScriptPlots3D(filename, plot_3D)
            if make_plots:
                runScript(filename)

    if not plots_only:
        # Write out stats marginalized
        mc.getMargeStats().saveAsText(rootdirname + '.margestats')

        # Limits from global likelihood
        if mc.loglikes is not None:
            mc.getLikeStats().saveAsText(rootdirname + '.likestats')

    # System command
    if finish_run_command:
        finish_run_command = finish_run_command.replace('%ROOTNAME%', rootname)
        os.system(finish_run_command)

    return "\n".join(result)


def make_param_file(file_name, feedback=True):
    with open(getdist.distparam_template, encoding="utf-8-sig") as f:
        content = f.read()
    with open(getdist.default_getdist_settings, encoding="utf-8-sig") as f:
        analysis = f.read()
    content = content.replace('%%%ANALYSIS_DEFAULTS%%%', analysis)
    with open(file_name, 'w', encoding="utf-8") as f:
        f.write(content)
    if feedback:
        print('Template .ini file written to ' + file_name)


def getdist_command(args=None):
    import argparse
    import getdist

    parser = argparse.ArgumentParser(description='GetDist sample analyser')
    parser.add_argument('ini_file', nargs='?',
                        help='.ini file with analysis settings (optional, if omitted uses defaults)')
    parser.add_argument('chain_root', nargs='?',
                        help='Root name of chain to analyse (e.g. chains/test), required unless file'
                             '_root specified in ini_file')
    parser.add_argument('--ignore_rows', type=float,
                        help='set initial fraction of chains to cut as burn in (fraction of total rows'
                             ', or >1 number of rows); overrides any value in ini_file if set')
    parser.add_argument('--make_param_file',
                        help='Produce a sample distparams.ini file that you can edit and use when running GetDist')
    parser.add_argument('--make_plots', action='store_true', help='Make PDFs from any requested plot script files')
    parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + getdist.__version__)
    args = parser.parse_args(args)
    if args.make_param_file:
        make_param_file(args.make_param_file)
    else:
        return getdist_script(args)


def getdist_gui():
    from getdist.gui.mainwindow import run_gui

    if sys.platform == "darwin":
        # On Mac need to run .app with plist to get menu name right (and avoid menu bugs)
        import subprocess
        import os

        path = os.path.join(os.path.dirname(getdist.gui.__file__), 'GetDist GUI.app')
        if os.path.exists(path):
            if subprocess.call(["/usr/bin/open", "-a", path, "--args"] + sys.argv[1:], env=os.environ):
                print("Error running 'GetDist GUI.app'. This may be a Catalina issue, any ideas?\n"
                      "Attempting to run script directly, using non-unified menus.")
                run_gui()
        else:
            print('GetDist GUI.app not found; not running getdist-gui, getdist package not installed '
                  'or no valid PySide2 found when setup was run. Running script...')
            run_gui()
    else:
        run_gui()



import decimal
import os
from io import BytesIO
import numpy as np
from getdist.paramnames import ParamInfo, ParamList
import tempfile

_sci_tolerance = 4


class TextFile:
    def __init__(self, lines=None):
        if isinstance(lines, str):
            lines = [lines]
        self.lines = lines or []

    def write(self, outfile):
        with open(outfile, 'w', encoding='utf-8') as f:
            f.write("\n".join(self.lines))


def texEscapeText(string):
    return string.replace('_', '{\\textunderscore}')


def times_ten_power(exponent):
    return r'\cdot 10^{%d}' % exponent


def float_to_decimal(f):
    # http://docs.python.org/library/decimal.html#decimal-faq
    """Convert a floating point number to a Decimal with no loss of information"""
    n, d = f.as_integer_ratio()
    numerator, denominator = decimal.Decimal(n), decimal.Decimal(d)
    ctx = decimal.Context(prec=60)
    result = ctx.divide(numerator, denominator)
    while ctx.flags[decimal.Inexact]:
        ctx.flags[decimal.Inexact] = False
        ctx.prec *= 2
        result = ctx.divide(numerator, denominator)
    return result


# noinspection PyUnboundLocalVariable
def numberFigs(number, sigfig, sci=False):
    # http://stackoverflow.com/questions/2663612/nicely-representing-a-floating-point-number-in-python/2663623#2663623
    assert (sigfig > 0)
    try:
        d = decimal.Decimal(number)
    except TypeError:
        d = float_to_decimal(float(number))
    if sci:
        exponent = d.adjusted()
        if abs(exponent) > _sci_tolerance:
            d = decimal.getcontext().multiply(d, float_to_decimal(10. ** -exponent))
        else:
            exponent = 0
    sign, digits = d.as_tuple()[0:2]
    if len(digits) < sigfig:
        digits = list(digits)
        digits.extend([0] * (sigfig - len(digits)))
    shift = d.adjusted()
    result = int(''.join(map(str, digits[:sigfig])))
    # Round the result
    if len(digits) > sigfig and digits[sigfig] >= 5:
        result += 1
    result = list(str(result))
    # Rounding can change the length of result
    # If so, adjust shift
    shift += len(result) - sigfig
    # reset len of result to sigfig
    result = result[:sigfig]
    if shift >= sigfig - 1:
        # Tack more zeros on the end
        result += ['0'] * (shift - sigfig + 1)
    elif 0 <= shift:
        # Place the decimal point in between digits
        result.insert(shift + 1, '.')
    else:
        # Tack zeros on the front
        assert (shift < 0)
        result = ['0.'] + ['0'] * (-shift - 1) + result
    if sign:
        result.insert(0, '-')
    if sci:
        return ''.join(result), exponent
    return ''.join(result)


class NumberFormatter:
    def __init__(self, sig_figs=4, separate_limit_tol=0.1, err_sf=2):
        self.sig_figs = sig_figs
        self.separate_limit_tol = separate_limit_tol
        self.err_sf = err_sf

    # noinspection PyUnboundLocalVariable
    def namesigFigs(self, value, limplus, limminus, wantSign=True, sci=False):
        frac = limplus / (abs(value) + limplus)
        sf = self.sig_figs
        if frac > 0.1 and 100 > value >= 20:
            sf = 2
        elif frac > 0.01 and value < 1000:
            sf = 3
        err_sf = self.err_sf
        if value >= 20 and frac > 0.1 and limplus >= 2:
            err_sf = 1
        if sci:
            # First, call without knowing sig figs, to get the exponent
            exponent = self.formatNumber(max(abs(value - limminus), abs(value + limplus)), sci=True)[1]
            if exponent:
                value, limplus, limminus = [
                    (lambda x: decimal.getcontext().multiply(
                        float_to_decimal(x), float_to_decimal(10. ** -exponent)))(lim)
                    for lim in [value, limplus, limminus]]
        plus_str = self.formatNumber(limplus, err_sf, wantSign)
        minus_str = self.formatNumber(limminus, err_sf, wantSign)
        res = self.formatNumber(value, sf)
        maxdp = max(self.decimal_places(plus_str), self.decimal_places(minus_str))
        while maxdp < self.decimal_places(res):
            sf -= 1
            if sf == 0:
                res = ('%.' + str(maxdp) + 'f') % value
                if float(res) == 0.0:
                    res = ('%.' + str(maxdp) + 'f') % 0
                break
            else:
                res = self.formatNumber(value, sf)

        while self.decimal_places(plus_str) > self.decimal_places(res):
            sf += 1
            res = self.formatNumber(value, sf)
        if sci:
            return res, plus_str, minus_str, exponent
        else:
            return res, plus_str, minus_str

    # noinspection PyUnboundLocalVariable
    def formatNumber(self, value, sig_figs=None, wantSign=False, sci=False):
        if sig_figs is None:
            sf = self.sig_figs
        else:
            sf = sig_figs
        s = numberFigs(value, sf, sci=sci)
        if sci:
            s, exponent = s
        if wantSign:
            if s[0] != '-' and float(s) < 0:
                s = '-' + s
            if float(s) > 0:
                s = '+' + s
        if sci:
            return s, exponent
        else:
            return s

    def decimal_places(self, s):
        i = s.find('.')
        if i > 0:
            return len(s) - i - 1
        return 0

    def plusMinusLimit(self, limit, upper, lower):
        return limit != 1 or abs(abs(upper / lower) - 1) > self.separate_limit_tol


class TableFormatter:
    def __init__(self):
        self.border = '|'
        self.endofrow = '\\\\'
        self.hline = '\\hline'
        self.paramText = 'Parameter'
        self.aboveTitles = self.hline
        self.majorDividor = '|'
        self.minorDividor = '|'
        self.colDividor = '||'
        self.belowTitles = ''
        self.headerWrapper = " %s"
        self.noConstraint = '---'
        self.spacer = ' '  # just to make output more readable
        self.colSeparator = self.spacer + '&' + self.spacer
        self.numberFormatter = NumberFormatter()

    def getLine(self, position=None):
        if position is not None and hasattr(self, position):
            return getattr(self, position)
        return self.hline

    def belowTitleLine(self, colsPerParam, numResults=None):
        return self.getLine("belowTitles")

    def startTable(self, ncol, colsPerResult, numResults):
        part = self.majorDividor + (" c" + self.minorDividor) * (colsPerResult - 1) + ' c'
        return '\\begin{tabular} {' + self.border + " l " + part * numResults + (
                self.colDividor + " l " + part * numResults) * (
                       ncol - 1) + self.border + '}'

    def endTable(self):
        return '\\end{tabular}'

    def titleSubColumn(self, colsPerResult, title):
        return ' \\multicolumn{' + str(
            colsPerResult) + '}{' + self.majorDividor + 'c' + self.majorDividor + '}{' + self.formatTitle(title) + '}'

    def formatTitle(self, title):
        return '\\bf ' + texEscapeText(title)

    def texEquation(self, txt):
        if txt and txt[0] != '$':
            return '$' + txt + '$'
        else:
            return txt

    def textAsColumn(self, txt, latex=False, separator=False, bold=False):
        wid = len(txt)
        if latex:
            wid += 2
            if bold:
                wid += 11
        res = txt + self.spacer * max(0, 28 - wid)
        if latex:
            res = self.texEquation(res)
            if bold:
                res = '{\\boldmath' + res + '}'
        if separator:
            res += self.colSeparator
        return res


class OpenTableFormatter(TableFormatter):
    def __init__(self):
        TableFormatter.__init__(self)
        self.border = ''
        self.aboveTitles = r'\noalign{\vskip 3pt}' + self.hline + r'\noalign{\vskip 1.5pt}' \
                           + self.hline + r'\noalign{\vskip 5pt}'
        self.belowTitles = r'\noalign{\vskip 3pt}' + self.hline
        self.aboveHeader = ''
        self.belowHeader = self.hline
        self.minorDividor = ''
        self.belowFinalRow = ''

    def titleSubColumn(self, colsPerResult, title):
        return ' \\multicolumn{' + str(colsPerResult) + '}{' + 'c' + '}{' + self.formatTitle(title) + '}'


class NoLineTableFormatter(OpenTableFormatter):
    def __init__(self):
        OpenTableFormatter.__init__(self)
        self.aboveHeader = ''
        # self.belowHeader = r'\noalign{\vskip 5pt}'
        self.minorDividor = ''
        self.majorDividor = ''
        self.belowFinalRow = self.hline
        self.belowBlockRow = self.hline
        self.colDividor = '|'
        self.hline = ''

    def belowTitleLine(self, colsPerParam, numResults=None):
        return r'\noalign{\vskip 3pt}\cline{2-' + str(colsPerParam * numResults + 1) + r'}\noalign{\vskip 3pt}'


class ResultTable:
    """
    Class for holding a latex table of parameter statistics
    """

    def __init__(self, ncol, results, limit=2, tableParamNames=None, titles=None, formatter=None,
                 numFormatter=None, blockEndParams=None, paramList=None, refResults=None, shiftSigma_indep=False,
                 shiftSigma_subset=False):
        """
        :param ncol: number of columns
        :param results: a :class:`MargeStats` or :class:`BestFit` instance, or a list of them for
                        comparing different results
        :param limit: which limit to include (1 is first limit calculated, usually 68%, 2 the second, usually 95%)
        :param tableParamNames: optional :class:`~.paramnames.ParamNames` instance listing particular
                                parameters to include
        :param titles: optional titles describing different results
        :param formatter: a table formatting class
        :param numFormatter: a number formatting class
        :param blockEndParams: mark parameters in blocks, ending on this list of parameter names
        :param paramList: a list of parameter names (strings) to include
        :param refResults: for showing parameter shifts, a reference :class:`MargeStats` instance to show differences to
        :param shiftSigma_indep: show parameter shifts in sigma assuming data are independent
        :param shiftSigma_subset: show parameter shifts in sigma assuming data are a subset of each other
        """
        # results is a margeStats or bestFit table
        self.lines = []
        if formatter is None:
            self.format = NoLineTableFormatter()
        else:
            self.format = formatter
        self.ncol = ncol
        if tableParamNames is None:
            self.tableParamNames = results[0]
        else:
            self.tableParamNames = tableParamNames
        if paramList is not None:
            self.tableParamNames = self.tableParamNames.filteredCopy(paramList)
        if numFormatter is not None:
            self.format.numFormatter = numFormatter

        self.results = results
        self.boldBaseParameters = True
        self.colsPerResult = len(results[0].getColumnLabels(limit))
        self.colsPerParam = len(results) * self.colsPerResult
        self.limit = limit
        self.refResults = refResults
        self.shiftSigma_indep = shiftSigma_indep
        self.shiftSigma_subset = shiftSigma_subset

        nparams = self.tableParamNames.numParams()
        numrow = nparams // ncol
        if nparams % ncol != 0:
            numrow += 1
        rows = []
        for par in self.tableParamNames.names[0:numrow]:
            rows.append([par])
        for col in range(1, ncol):
            for i in range(numrow * col, min(numrow * (col + 1), nparams)):
                rows[i - numrow * col].append(self.tableParamNames.names[i])

        self.lines.append(self.format.startTable(ncol, self.colsPerResult, len(results)))
        if titles is not None:
            self.addTitlesRow(titles)
        self.addHeaderRow()
        for row in rows[:-1]:
            self.addFullTableRow(row)
            if ncol == 1 and blockEndParams is not None and row[0].name in blockEndParams:
                self.addLine("belowBlockRow")
            else:
                self.addLine("belowRow")
        self.addFullTableRow(rows[-1])
        self.addLine("belowFinalRow")
        self.endTable()

    def addFullTableRow(self, row):
        txt = self.format.colSeparator.join(self.paramLabelColumn(param) + self.paramResultsTex(param) for param in row)
        if not self.ncol == len(row):
            txt += self.format.colSeparator * ((1 + self.colsPerParam) * (self.ncol - len(row)))
        self.lines.append(txt + self.format.endofrow)

    def addLine(self, position):
        if self.format.getLine(position) is None:  # no line is appended if the attribute is None
            return self.lines
        else:
            return self.lines.append(self.format.getLine(position))

    def addTitlesRow(self, titles):
        self.addLine("aboveTitles")
        cols = [self.format.titleSubColumn(1, '')]
        cols += [self.format.titleSubColumn(self.colsPerResult, title) for title in titles]
        self.lines.append(self.format.colSeparator.join(cols * self.ncol) + self.format.endofrow)

        belowTitleLine = self.format.belowTitleLine(self.colsPerResult, self.colsPerParam // self.colsPerResult)
        if belowTitleLine:
            self.lines.append(belowTitleLine)

    def addHeaderRow(self):
        self.addLine("aboveHeader")
        cols = [self.format.headerWrapper % self.format.paramText]
        for result in self.results:
            cols += [self.format.headerWrapper % s for s in result.getColumnLabels(self.limit)]
        self.lines.append(self.format.colSeparator.join(cols * self.ncol) + self.format.endofrow)

        self.addLine("belowHeader")

    def paramResultsTex(self, param):
        return self.format.colSeparator.join(self.paramResultTex(result, param) for result in self.results)

    def paramResultTex(self, result, p):
        values = result.texValues(self.format, p, self.limit, self.refResults,
                                  shiftSigma_subset=self.shiftSigma_subset, shiftSigma_indep=self.shiftSigma_indep)
        if values is not None:
            if len(values) > 1:
                txt = self.format.textAsColumn(values[1], True, separator=True)
            else:
                txt = ''
            txt += self.format.textAsColumn(values[0], values[0] != self.format.noConstraint)
            return txt
        else:
            return self.format.textAsColumn('') * len(result.getColumnLabels(self.limit))

    def paramLabelColumn(self, param):
        return self.format.textAsColumn(param.getLabel(), True, separator=True, bold=not param.isDerived)

    def endTable(self):
        self.lines.append(self.format.endTable())

    def tableTex(self, document=False, latex_preamble=None, packages=('amsmath', 'amssymb', 'bm')):
        """
        Get the latex string for the table

        :param document: if True, make a full latex file, if False just the snippet for including in another file
        :param latex_preamble: any preamble to include in the latex file
        :param packages: list of packages to load
        """

        if document:
            lines = [r'\documentclass{article}', r'\pagestyle{empty}']
            for package in packages:
                lines.append(r'\usepackage{%s}' % package)
            lines.append('\\renewcommand{\\arraystretch}{1.5}')
            if latex_preamble:
                lines.append(latex_preamble)
            lines.append('\\begin{document}')
            lines += self.lines
            lines.append('\\end{document}')
        else:
            lines = self.lines
        return "\n".join(lines)

    def write(self, fname, **kwargs):
        """
        Write the latex for the table to a file

        :param fname: filename to write
        :param kwargs: arguments for :func:`~ResultTable.tableTex`
        """
        TextFile(self.tableTex(**kwargs)).write(fname)

    def tablePNG(self, dpi=None, latex_preamble=None, filename=None, bytesIO=False):
        """
        Get a .png file image of the table. You must have latex installed to use this.

        :param dpi: dpi settings for the png
        :param latex_preamble: any latex preamble
        :param filename: filename to save to (defaults to file in the temp directory)
        :param bytesIO: if True, return a BytesIO instance holding the .png data
        :return: if bytesIO, the BytesIO instance, otherwise name of the output file
        """
        texfile = tempfile.mktemp(suffix='.tex')
        self.write(texfile, document=True, latex_preamble=latex_preamble)
        basefile = os.path.splitext(texfile)[0]
        outfile = filename or basefile + '.png'
        old_pwd = os.getcwd()

        def runCommand(command):
            command += ' 2>%s 1>&2' % os.devnull
            os.system(command)

        try:
            os.chdir(os.path.dirname(texfile))
            runCommand('latex %s' % texfile)
            cmd = 'dvipng'
            if dpi:
                cmd += ' -D %s' % dpi
            cmd += ' -T tight -x 1000 -z 9 --truecolor -o "%s" "%s" ' \
                   % (outfile, basefile + '.dvi')
            runCommand(cmd)
        finally:
            for f in [basefile + ext for ext in ('.tex', '.dvi', '.aux', '.log')]:
                if os.path.isfile(f):
                    os.remove(f)
            os.chdir(old_pwd)
        if bytesIO:
            with open(outfile, 'rb') as f:
                result = BytesIO(f.read())
            os.remove(outfile)
            result.seek(0)
            return result
        else:
            return outfile


class ParamResults(ParamList):
    """
    Base class for a set of parameter results, inheriting from :class:`~.paramnames.ParamList`,
    so that self.names is a list of :class:`~.paramnames.ParamInfo` instances for each parameter, which
    have attribute holding results for the different parameters.
    """
    pass


class LikelihoodChi2:
    name: str
    tag: str
    chisq: float


class BestFit(ParamResults):
    """
    Class holding the result of a likelihood minimization, inheriting from :class:`ParamResults`.
    The data is read from a specific formatted text file (.minimum or .bestfit) as output by CosmoMC or Cobaya.
    """

    def __init__(self, fileName=None, setParamNameFile=None, want_fixed=False, max_posterior=True):
        """
        :param fileName: text file to load from, assumed to be in CosmoMC's .minimum format
        :param setParamNameFile: optional name of .paramnames file listing preferred parameter labels for the parameters
        :param want_fixed:  whether to include values of parameters that are not allowed to vary
        :param max_posterior: whether the file is a maximum posterior (default) or maximum likelihood
        """

        ParamResults.__init__(self)
        self.max_posterior = max_posterior
        if fileName is not None:
            self.loadFromFile(fileName, want_fixed=want_fixed)
        if setParamNameFile is not None:
            self.setLabelsFromParamNames(setParamNameFile)

    def getColumnLabels(self, **kwargs):
        return ['Best fit']

    def loadFromFile(self, filename, want_fixed=False):
        textFileLines = self.fileList(filename)
        first = textFileLines[0].strip().split('=')
        if first[0].strip() == 'weight':
            self.weight = float(first[1].strip())
            del (textFileLines[0])
            first = textFileLines[0].strip().split('=')
        if first[0].strip() != '-log(Like)':
            raise Exception('Error in format of parameter (best fit) file')
        self.logLike = float(first[1].strip())
        isFixed = False
        isDerived = False
        self.chiSquareds = []
        chunks = 0
        if len(textFileLines[1].strip()) > 0:
            del (textFileLines[1])  # if it has chi2 line as well
        for ix in range(2, len(textFileLines)):
            line = textFileLines[ix]
            if len(line.strip()) == 0:
                chunks += 1
                isFixed = not isFixed
                isDerived = True
                if chunks == 3:
                    if ix + 2 >= len(textFileLines):
                        break
                    for likePart in textFileLines[ix + 2:]:
                        if len(likePart.strip()) != 0:
                            (chisq, name) = [s.strip() for s in likePart.split(None, 2)][1:]
                            name = [s.strip() for s in name.split(':', 1)]
                            if len(name) > 1:
                                (kind, name) = name
                            else:
                                kind = ''
                            chi2 = LikelihoodChi2()
                            if '=' in name:
                                chi2.tag, chi2.name = [s.strip() for s in name.split('=')]
                            else:
                                chi2.tag, chi2.name = None, name
                            chi2.chisq = float(chisq)
                            self.chiSquareds.append((kind, chi2))
                    break
                continue
            if not isFixed or want_fixed:
                param = ParamInfo()
                param.isFixed = isFixed
                param.isDerived = isDerived
                (param.number, param.best_fit, param.name, param.label) = [s.strip() for s in line.split(None, 3)]
                param.number = int(param.number)
                param.best_fit = float(param.best_fit)
                self.names.append(param)

    def sortedChiSquareds(self):
        likes = dict()
        for kind, val in self.chiSquareds:
            if kind not in likes:
                likes[kind] = []
            likes[kind].append(val)
        return sorted(iter(likes.items()))

    def chiSquareForKindName(self, kind, name):
        for akind, val in self.chiSquareds:
            if akind == kind and val.name == name:
                return val.chisq
        return None

    def texValues(self, formatter, p, **kwargs):
        param = self.parWithName(p.name)
        if param is not None:
            return [formatter.numberFormatter.formatNumber(param.best_fit)]
        else:
            return None

    def getParamDict(self, include_derived=True):
        res = dict()
        for i, name in enumerate(self.names):
            if include_derived or not name.isDerived:
                res[name.name] = name.best_fit
        res['weight'] = 1
        res['loglike'] = self.logLike
        return res


class ParamLimit:
    """
    Class containing information about a marginalized parameter limit.

    :ivar lower: lower limit
    :ivar upper: upper limit
    :ivar twotail: True if a two-tail limit, False if one-tail
    :ivar onetail_upper: True if one-tail upper limit
    :ivar onetail_lower: True if one-tail lower limit
    """

    def __init__(self, minmax, tag='two'):
        """
        :param minmax: a [min,max] tuple with lower and upper limits. Entries be None if no limit.
        :param tag: a text tag descibing the limit, one of ['two' | '>' | '<' | 'none']
        """

        self.lower = minmax[0]
        self.upper = minmax[1]
        self.twotail = tag == 'two'
        self.onetail_upper = tag == '>'
        self.onetail_lower = tag == '<'

    def limitTag(self):
        """
        :return: Short text tag describing the type of limit (one-tail or two tail):

                - *two*: two-tail limit
                - *>*: a one-tail upper limit
                - *<*: a one-tail lower limit
                - *none*: no limits (both boundaries have high probability)
        """
        if self.twotail:
            return 'two'
        elif self.onetail_upper:
            return '>'
        elif self.onetail_lower:
            return '<'
        else:
            return 'none'

    def limitType(self):
        """
        :return: a text description of the type of limit. One of:

            - *two tail*
            - *one tail upper limit*
            - *one tail lower limit*
            - *none*
        """
        if self.twotail:
            return 'two tail'
        elif self.onetail_upper:
            return 'one tail upper limit'
        elif self.onetail_lower:
            return 'one tail lower limit'
        else:
            return 'none'

    def __str__(self):
        """
        :return: string representation of lower and upper bounds, with text description of the limit type
        """
        return "%g %g %s" % (self.lower, self.upper, self.limitTag())


class MargeStats(ParamResults):
    """
    Stores marginalized 1D parameter statistics, including mean, variance and confidence limits,
    inheriting from :class:`ParamResults`.

    Values are stored as attributes of the :class:`~.paramnames.ParamInfo` objects stored in self.names.
    Use *par= margeStats.parWithName('xxx')* to get the :class:`~.paramnames.ParamInfo` for parameter *xxx*;
    Values stored are:

    - *par.mean*: parameter mean
    - *par.err*: standard deviation
    - *limits*: list of :class:`~.types.ParamLimit` objects for the stored number of marginalized limits

    For example to get the first and second lower limits (default 68% and 95%) for parameter *xxx*::

         print(margeStats.names.parWithName('xxx').limits[0].lower)
         print(margeStats.names.parWithName('xxx').limits[1].lower)

    See  :class:`~.types.ParamLimit` for details of limits.
    """

    def loadFromFile(self, filename):
        """
        Load from a plain text file

        :param filename: file to load from
        """
        textFileLines = self.fileList(filename)
        lims = textFileLines[0].split(':')[1]
        self.limits = [float(s.strip()) for s in lims.split(';')]
        self.hasBestFit = False
        for line in textFileLines[3:]:
            if len(line.strip()) == 0:
                break
            param = ParamInfo()
            items = [s.strip() for s in line.split(None, len(self.limits) * 3 + 3)]
            param.name = items[0]
            if param.name[-1] == '*':
                param.isDerived = True
                param.name = param.name[:-1]
            param.mean = float(items[1])
            param.err = float(items[2])
            param.label = items[-1]
            param.limits = []
            for i in range(len(self.limits)):
                param.limits.append(ParamLimit([float(s) for s in items[3 + i * 3:5 + i * 3]], items[5 + i * 3]))
            self.names.append(param)

    def headerLine(self, inc_limits=False):
        parForm = self.parFormat()
        text = ""
        text += parForm % "parameter" + "  "
        text += "%-15s" % "mean"
        text += "%-15s" % "sddev"
        for j, limit in enumerate(self.limits):
            if inc_limits:
                tag = "_%.0f%%" % (limit * 100)
                limtxt = 'type'
            else:
                tag = str(j + 1)
                limtxt = "limit" + tag
            text += "%-15s" % ("lower" + tag)
            text += "%-15s" % ("upper" + tag)
            text += "%-7s" % limtxt
        return text, parForm

    def __str__(self):
        contours_str = '; '.join([str(c) for c in self.limits])
        header, parForm = self.headerLine()
        text = ""
        text += "Marginalized limits: %s\n\n" % contours_str
        text += header
        text += "\n"

        for j, par in enumerate(self.names):
            text += parForm % (self.name(j, True))
            text += "%15.7E%15.7E" % (par.mean, par.err)
            for lim in par.limits:
                text += "%15.7E%15.7E  %-5s" % (lim.lower, lim.upper, lim.limitTag())
            text += "   %s\n" % par.label
        return text

    def addBestFit(self, bf):
        self.hasBestFit = True
        self.logLike = bf.logLike
        # the next line deletes parameters not in best-fit;
        # this is good e.g. to get rid of yhe from importance sampled result
        self.names = [x for x in self.names if bf.parWithName(x.name) is not None]
        for par in self.names:
            param = bf.parWithName(par.name)
            par.best_fit = param.best_fit
            par.isDerived = param.isDerived

    def limitText(self, limit):
        txt = str(round(self.limits[limit - 1] * 100.))
        if txt.endswith(".0"):  # e.g. 95.0 -> 95
            txt = txt.split(".")[0]
        return txt

    def getColumnLabels(self, limit=2):
        if self.hasBestFit:
            res = ['Best fit']
        else:
            res = []
        return res + [self.limitText(limit) + '\\% limits']

    def texValues(self, formatter, p, limit=2, refResults=None, shiftSigma_indep=False, shiftSigma_subset=False):
        if not isinstance(p, ParamInfo):
            param = self.parWithName(p)
        else:
            param = self.parWithName(p.name)
        if param is not None:
            lim = param.limits[limit - 1]
            sf = 3
            if param.name.startswith('chi2'):
                # Chi2 for low dof are very skewed, always want mean and sigma or limit
                res, sigma, _ = formatter.numberFormatter.namesigFigs(param.mean, param.err, param.err, wantSign=False,
                                                                      sci=False)
                if limit == 1:
                    res += r'\pm ' + sigma
                else:
                    # in this case give mean and effective dof
                    res += r'\,({\nu\rm{:}\,%.1f})' % (param.err ** 2 / 2)
                    # res, plus_str, minus_str =
                    # formatter.numberFormatter.namesigFigs(param.mean, lim.upper - param.mean, lim.lower, sci=False)
                    # res += '^{' + plus_str + '}_{>' + minus_str + '}'
            elif lim.twotail:
                if not formatter.numberFormatter.plusMinusLimit(limit, lim.upper - param.mean, lim.lower - param.mean):
                    res, plus_str, _, exponent = formatter.numberFormatter.namesigFigs(param.mean, param.err, param.err,
                                                                                       wantSign=False, sci=True)
                    res += r'\pm ' + plus_str
                else:
                    res, plus_str, minus_str, exponent = formatter.numberFormatter.namesigFigs(param.mean,
                                                                                               lim.upper - param.mean,
                                                                                               lim.lower - param.mean,
                                                                                               sci=True)
                    res += '^{' + plus_str + '}_{' + minus_str + '}'
                if exponent:
                    res = r'\left(\,%s\,\right)' % res + times_ten_power(exponent)
            elif lim.onetail_upper:
                res, exponent = formatter.numberFormatter.formatNumber(lim.upper, sf, sci=True)
                res = '< ' + res
                if exponent:
                    res += times_ten_power(exponent)
            elif lim.onetail_lower:
                res, exponent = formatter.numberFormatter.formatNumber(lim.lower, sf, sci=True)
                res = '> ' + res
                if exponent:
                    res += times_ten_power(exponent)
            else:
                res = formatter.noConstraint
            if refResults is not None and res != formatter.noConstraint:
                refVal = refResults.parWithName(param.name)
                if refVal is not None:
                    delta = param.mean - refVal.mean
                    if shiftSigma_indep or shiftSigma_subset:
                        res += r'\quad('
                        if shiftSigma_subset:
                            # give mean shift in sigma units for subset data (regularized to max sigma/20)
                            subset_sigma = np.sqrt(abs(param.err ** 2 - refVal.err ** 2))
                            res += '%+.1f \\sigma_s' % (delta / max(subset_sigma, refVal.err / 20))
                        if shiftSigma_indep:
                            # give mean shift in sigma units for independent data
                            indep_sigma = np.sqrt(param.err ** 2 + refVal.err ** 2)
                            res += ', %+.1f \\sigma_i' % (delta / indep_sigma)
                        res += ')'
                    else:
                        res += r'\quad(%+.1f \\sigma)' % (delta / refVal.err)
            if self.hasBestFit:  # add best fit too
                rangew = (lim.upper - lim.lower) / 10
                bestfit, _, _, exponent = formatter.numberFormatter.namesigFigs(param.best_fit, rangew, -rangew,
                                                                                sci=True)
                if exponent:
                    bestfit += times_ten_power(exponent)
                return [res, bestfit]
            return [res]
        else:
            return None


class LikeStats(ParamResults):
    """
    Stores likelihood-related statistics, including best-fit sample and extremal values of the N-D confidence region,
    inheriting from :class:`ParamResults`.
    TODO: currently only saves to text, does not load full data from file
    """

    def loadFromFile(self, filename):
        textFileLines = self.fileList(filename)
        results = dict()
        for line in textFileLines:
            if len(line.strip()) == 0:
                break
            name, value = [x.strip() for x in line.split('=')]
            results[name] = float(value)
        self.logLike_sample = results.get('Best fit sample -log(Like)', None)
        self.logMeanInvLike = results.get('Ln(mean 1/like)', None)
        self.meanLogLike = results.get('mean(-Ln(like))', None)
        self.logMeanLike = results.get('-Ln(mean like)', None)
        self.complexity = results.get('complexity', None)

        # TODO: load N-D limits

    def likeSummary(self):
        text = "Best fit sample -log(Like) = %f\n" % self.logLike_sample
        if self.logMeanInvLike:
            text += "Ln(mean 1/like) = %f\n" % self.logMeanInvLike
        text += "mean(-Ln(like)) = %f\n" % self.meanLogLike
        text += "-Ln(mean like)  = %f\n" % self.logMeanLike
        # text += "complexity = %f\n" % self.complexity

        return text

    def headerLine(self):
        return self.parFormat() % "parameter" + '  bestfit        lower1         upper1         lower2         upper2\n'

    def __str__(self):
        text = self.likeSummary()
        parForm = self.parFormat()
        if len(self.names):
            text += "\n"
            text += self.headerLine()
            for j, par in enumerate(self.names):
                text += parForm % (self.name(j, True))
                text += '%15.7E%15.7E%15.7E%15.7E%15.7E   %s\n' % (par.bestfit_sample,
                                                                   par.ND_limit_bot[0], par.ND_limit_top[0],
                                                                   par.ND_limit_bot[1], par.ND_limit_top[1], par.label)
        return text


class ConvergeStats(ParamResults):
    def loadFromFile(self, filename):
        try:
            textFileLines = self.fileList(filename)
            self.R_eigs = []
            for i in range(len(textFileLines)):
                if textFileLines[i].find('var(mean)') >= 0:
                    for line in textFileLines[i + 1:]:
                        if len(line.strip()) == 0:
                            break
                        try:
                            self.R_eigs.append(line.split()[1])
                        except Exception:
                            self.R_eigs.append('1e30')
                elif 'Parameter auto-correlations' in textFileLines[i]:
                    self.auto_correlation_steps = [int(s) for s in textFileLines[i + 2].split()]
                    self.auto_correlations = []
                    self.auto_correlation_pars = []
                    for line in textFileLines[i + 3:]:
                        if len(line.strip()) == 0:
                            break
                        items = line.split(None, len(self.auto_correlation_steps) + 1)
                        self.auto_correlation_pars.append(items[0])
                        self.auto_correlations.append([float(s) for s in items[1:-1]])
        except:
            print('Error reading: ' + filename)
            raise

    def worstR(self):
        return self.R_eigs[len(self.R_eigs) - 1]



import sys
import fnmatch
import os
from getdist import covmat

if len(sys.argv) < 4:
    print('covscale rescales parameter(s) in all .covmat files in a directory and outputs to another directory')
    print('Usage: python covscale.py in_dir out_dir param1:param2:.. fac1:fac2:..')
    sys.exit()

indir = os.path.abspath(sys.argv[1]) + os.sep
outdir = os.path.abspath(sys.argv[2]) + os.sep
pars = sys.argv[3].split(':')
factors = sys.argv[4].split(':')

if not os.path.exists(outdir):
    os.makedirs(outdir)

for f in os.listdir(indir):
    if fnmatch.fnmatch(f, "*.covmat"):
        print(indir + f)
        cov = covmat.CovMat(indir + f)
        for par, factor in zip(pars, factors):
            cov.rescaleParameter(par, float(factor))
        cov.saveToFile(outdir + f)
